# Copyright 2014 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from unittest.mock import call
from unittest.mock import Mock
from unittest.mock import patch
from sqlalchemy.engine import reflection
from sqlalchemy.schema import Column

from trove.db.sqlalchemy.migrate_repo.schema import String
from trove.db.sqlalchemy import utils as db_utils
from trove.tests.unittests import trove_testtools


class TestDbMigrationUtils(trove_testtools.TestCase):
    def setUp(self):
        super(TestDbMigrationUtils, self).setUp()

    def tearDown(self):
        super(TestDbMigrationUtils, self).tearDown()

    @patch.object(reflection.Inspector, 'from_engine')
    def test_get_foreign_key_constraint_names_single_match(self,
                                                           mock_inspector):
        mock_engine = Mock()
        (mock_inspector.return_value.
         get_foreign_keys.return_value) = [{'constrained_columns': ['col1'],
                                            'referred_table': 'ref_table1',
                                            'referred_columns': ['ref_col1'],
                                            'name': 'constraint1'},
                                           {'constrained_columns': ['col2'],
                                            'referred_table': 'ref_table2',
                                            'referred_columns': ['ref_col2'],
                                            'name': 'constraint2'}]
        ret_val = db_utils.get_foreign_key_constraint_names(mock_engine,
                                                            'table1',
                                                            ['col1'],
                                                            'ref_table1',
                                                            ['ref_col1'])
        self.assertEqual(['constraint1'], ret_val)

    @patch.object(reflection.Inspector, 'from_engine')
    def test_get_foreign_key_constraint_names_multi_match(self,
                                                          mock_inspector):
        mock_engine = Mock()
        (mock_inspector.return_value.
         get_foreign_keys.return_value) = [
             {'constrained_columns': ['col1'],
              'referred_table': 'ref_table1',
              'referred_columns': ['ref_col1'],
              'name': 'constraint1'},
             {'constrained_columns': ['col2', 'col3'],
              'referred_table': 'ref_table1',
              'referred_columns': ['ref_col2', 'ref_col3'],
              'name': 'constraint2'},
             {'constrained_columns': ['col2', 'col3'],
              'referred_table': 'ref_table1',
              'referred_columns': ['ref_col2', 'ref_col3'],
              'name': 'constraint3'},
             {'constrained_columns': ['col4'],
              'referred_table': 'ref_table2',
              'referred_columns': ['ref_col4'],
              'name': 'constraint4'}]
        ret_val = db_utils.get_foreign_key_constraint_names(
            mock_engine, 'table1', ['col2', 'col3'],
            'ref_table1', ['ref_col2', 'ref_col3'])
        self.assertEqual(['constraint2', 'constraint3'], ret_val)

    @patch.object(reflection.Inspector, 'from_engine')
    def test_get_foreign_key_constraint_names_no_match(self, mock_inspector):
        mock_engine = Mock()
        (mock_inspector.return_value.
         get_foreign_keys.return_value) = []
        ret_val = db_utils.get_foreign_key_constraint_names(mock_engine,
                                                            'table1',
                                                            ['col1'],
                                                            'ref_table1',
                                                            ['ref_col1'])
        self.assertEqual([], ret_val)

    @patch('trove.db.sqlalchemy.utils.ForeignKeyConstraint')
    def test_drop_foreign_key_constraints(self, mock_constraint):
        test_columns = [Column('col1', String(5)),
                        Column('col2', String(5))]
        test_refcolumns = [Column('ref_col1', String(5)),
                           Column('ref_col2', String(5))]
        test_constraint_names = ['constraint1', 'constraint2']
        db_utils.drop_foreign_key_constraints(test_constraint_names,
                                              test_columns,
                                              test_refcolumns)
        expected = [call(columns=test_columns,
                         refcolumns=test_refcolumns,
                         name='constraint1'),
                    call(columns=test_columns,
                         refcolumns=test_refcolumns,
                         name='constraint2')]
        self.assertEqual(expected, mock_constraint.call_args_list)



# Copyright 2014 Rackspace
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest.mock import Mock
from unittest.mock import patch

from trove.common import exception
from trove.datastore import models as datastore_models
from trove.datastore.models import Datastore
from trove.tests.unittests.datastore.base import TestDatastoreBase


class TestDatastore(TestDatastoreBase):

    def test_create_failure_with_datastore_default_not_defined(self):
        self.assertRaises(
            exception.DatastoreDefaultDatastoreNotDefined,
            datastore_models.get_datastore_version)

    def test_load_datastore(self):
        datastore = Datastore.load(self.ds_name)
        self.assertEqual(self.ds_name, datastore.name)

    @patch.object(datastore_models, 'CONF')
    def test_create_failure_with_datastore_default(self, mock_conf):
        mock_conf.default_datastore = 'bad_ds'
        self.assertRaisesRegex(exception.DatastoreDefaultDatastoreNotFound,
                               "Default datastore 'bad_ds' cannot be found",
                               datastore_models.get_datastore_version)
        self.assertRaisesRegex(exception.DatastoreNotFound,
                               "Datastore 'my_ds' cannot be found",
                               datastore_models.get_datastore_version,
                               'my_ds')

    def test_get_datastore_or_version(self):
        # datastore, datastore_version, valid, exception
        data = [
            [None, None, True],
            ['ds', None, True],
            ['ds', 'ds_ver', True],
            [None, 'ds_ver', False, exception.DatastoreNoVersion],
        ]
        for datum in data:
            ds_id = datum[0]
            ds_ver_id = datum[1]
            valid = datum[2]
            expected_exception = None
            if not valid:
                expected_exception = datum[3]
            ds = Mock()
            ds.id = ds_id
            ds.name = ds_id
            ds_ver = Mock()
            ds_ver.id = ds_ver_id
            ds_ver.name = ds_ver_id
            ds_ver.datastore_id = ds_id
            with patch.object(datastore_models.Datastore, 'load',
                              return_value=ds):
                with patch.object(datastore_models.DatastoreVersion, 'load',
                                  return_value=ds_ver):
                    if valid:
                        (get_ds_id, get_ds_ver_id) = (
                            datastore_models.get_datastore_or_version(
                                ds_id, ds_ver_id))
                        self.assertEqual(ds_id, get_ds_id)
                        self.assertEqual(ds_ver_id, get_ds_ver_id)
                    else:
                        self.assertRaises(
                            expected_exception,
                            datastore_models.get_datastore_or_version,
                            ds_id, ds_ver_id)



#!/usr/bin/env python

# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Colorizer Code is borrowed from Twisted:
# Copyright (c) 2001-2010 Twisted Matrix Laboratories.
#
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files (the
#    "Software"), to deal in the Software without restriction, including
#    without limitation the rights to use, copy, modify, merge, publish,
#    distribute, sublicense, and/or sell copies of the Software, and to
#    permit persons to whom the Software is furnished to do so, subject to
#    the following conditions:
#
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
#    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
#    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
#    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""Unittest runner for Nova.

To run all tests
    python run_tests.py

To run a single test:
    python run_tests.py test_compute:ComputeTestCase.test_run_terminate

To run a single test module:
    python run_tests.py test_compute

    or

    python run_tests.py api.test_wsgi

"""

import gettext
import heapq
import logging
import os
import unittest
import sys
import time

gettext.install('nova')

from nose import config
from nose import core
from nose import result
from proboscis import case
from proboscis import SkipTest

class _AnsiColorizer(object):
    """
    A colorizer is an object that loosely wraps around a stream, allowing
    callers to write text to the stream in a particular color.

    Colorizer classes must implement C{supported()} and C{write(text, color)}.
    """
    _colors = dict(black=30, red=31, green=32, yellow=33,
                   blue=34, magenta=35, cyan=36, white=37)

    def __init__(self, stream):
        self.stream = stream

    def supported(cls, stream=sys.stdout):
        """
        A class method that returns True if the current platform supports
        coloring terminal output using this method. Returns False otherwise.
        """
        if not stream.isatty():
            return False  # auto color only on TTYs
        try:
            import curses
        except ImportError:
            return False
        else:
            try:
                try:
                    return curses.tigetnum("colors") > 2
                except curses.error:
                    curses.setupterm()
                    return curses.tigetnum("colors") > 2
            except:
                raise
                # guess false in case of error
                return False
    supported = classmethod(supported)

    def write(self, text, color):
        """
        Write the given text to the stream in the given color.

        @param text: Text to be written to the stream.

        @param color: A string label for a color. e.g. 'red', 'white'.
        """
        color = self._colors[color]
        self.stream.write('\x1b[%s;1m%s\x1b[0m' % (color, text))


class _Win32Colorizer(object):
    """
    See _AnsiColorizer docstring.
    """
    def __init__(self, stream):
        from win32console import GetStdHandle, STD_OUT_HANDLE, \
             FOREGROUND_RED, FOREGROUND_BLUE, FOREGROUND_GREEN, \
             FOREGROUND_INTENSITY
        red, green, blue, bold = (FOREGROUND_RED, FOREGROUND_GREEN,
                                  FOREGROUND_BLUE, FOREGROUND_INTENSITY)
        self.stream = stream
        self.screenBuffer = GetStdHandle(STD_OUT_HANDLE)
        self._colors = {
            'normal': red | green | blue,
            'red': red | bold,
            'green': green | bold,
            'blue': blue | bold,
            'yellow': red | green | bold,
            'magenta': red | blue | bold,
            'cyan': green | blue | bold,
            'white': red | green | blue | bold
            }

    def supported(cls, stream=sys.stdout):
        try:
            import win32console
            screenBuffer = win32console.GetStdHandle(
                win32console.STD_OUT_HANDLE)
        except ImportError:
            return False
        import pywintypes
        try:
            screenBuffer.SetConsoleTextAttribute(
                win32console.FOREGROUND_RED |
                win32console.FOREGROUND_GREEN |
                win32console.FOREGROUND_BLUE)
        except pywintypes.error:
            return False
        else:
            return True
    supported = classmethod(supported)

    def write(self, text, color):
        color = self._colors[color]
        self.screenBuffer.SetConsoleTextAttribute(color)
        self.stream.write(text)
        self.screenBuffer.SetConsoleTextAttribute(self._colors['normal'])


class _NullColorizer(object):
    """
    See _AnsiColorizer docstring.
    """
    def __init__(self, stream):
        self.stream = stream

    def supported(cls, stream=sys.stdout):
        return True
    supported = classmethod(supported)

    def write(self, text, color):
        self.stream.write(text)


def get_elapsed_time_color(elapsed_time):
    if elapsed_time > 1.0:
        return 'yellow'
    elif elapsed_time > 0.25:
        return 'cyan'
    else:
        return 'green'


class NovaTestResult(case.TestResult):
    def __init__(self, *args, **kw):
        self.show_elapsed = kw.pop('show_elapsed')
        self.known_bugs = kw.pop('known_bugs', {})
        super(NovaTestResult, self).__init__(*args, **kw)
        self.num_slow_tests = 5
        self.slow_tests = []  # this is a fixed-sized heap
        self._last_case = None
        self.colorizer = None
        # NOTE(vish): reset stdout for the terminal check
        stdout = sys.stdout
        sys.stdout = sys.__stdout__
        for colorizer in [_Win32Colorizer, _AnsiColorizer, _NullColorizer]:
            if colorizer.supported():
                self.colorizer = colorizer(self.stream)
                break
        sys.stdout = stdout

        # NOTE(lorinh): Initialize start_time in case a sqlalchemy-migrate
        # error results in it failing to be initialized later. Otherwise,
        # _handleElapsedTime will fail, causing the wrong error message to
        # be outputted.
        self.start_time = time.time()

    def _intercept_known_bugs(self, test, err):
        name = str(test)
        excuse = self.known_bugs.get(name, None)
        if excuse:
            tracker_id, error_string = excuse
            if error_string in str(err[1]):
                skip = SkipTest("KNOWN BUG: %s\n%s"
                                % (tracker_id, str(err[1])))
                self.onError(test)
                super(NovaTestResult, self).addSkip(test, skip)
            else:
                result = (RuntimeError, RuntimeError(
                     'Test "%s" contains known bug %s.\n'
                     'Expected the following error string:\n%s\n'
                     'What was seen was the following:\n%s\n'
                     'If the bug is no longer happening, please change '
                     'the test config.'
                     % (name, tracker_id, error_string, str(err))), None)
                self.onError(test)
                super(NovaTestResult, self).addError(test, result)
            return True
        return False

    def getDescription(self, test):
        return str(test)

    def _handleElapsedTime(self, test):
        self.elapsed_time = time.time() - self.start_time
        item = (self.elapsed_time, test)
        # Record only the n-slowest tests using heap
        if len(self.slow_tests) >= self.num_slow_tests:
            heapq.heappushpop(self.slow_tests, item)
        else:
            heapq.heappush(self.slow_tests, item)

    def _writeElapsedTime(self, test):
        color = get_elapsed_time_color(self.elapsed_time)
        self.colorizer.write("  %.2f" % self.elapsed_time, color)

    def _writeResult(self, test, long_result, color, short_result, success):
        if self.showAll:
            self.colorizer.write(long_result, color)
            if self.show_elapsed and success:
                self._writeElapsedTime(test)
            self.stream.writeln()
        elif self.dots:
            self.stream.write(short_result)
            self.stream.flush()

    # NOTE(vish): copied from unittest with edit to add color
    def addSuccess(self, test):
        if self._intercept_known_bugs(test, None):
            return
        unittest.TestResult.addSuccess(self, test)
        self._handleElapsedTime(test)
        self._writeResult(test, 'OK', 'green', '.', True)

    # NOTE(vish): copied from unittest with edit to add color
    def addFailure(self, test, err):
        if self._intercept_known_bugs(test, err):
            return
        self.onError(test)
        unittest.TestResult.addFailure(self, test, err)
        self._handleElapsedTime(test)
        self._writeResult(test, 'FAIL', 'red', 'F', False)

    # NOTE(vish): copied from nose with edit to add color
    def addError(self, test, err):
        """Overrides normal addError to add support for
        errorClasses. If the exception is a registered class, the
        error will be added to the list for that class, not errors.
        """
        if self._intercept_known_bugs(test, err):
            return
        self.onError(test)
        self._handleElapsedTime(test)
        stream = getattr(self, 'stream', None)
        ec, ev, tb = err
        try:
            exc_info = self._exc_info_to_string(err, test)
        except TypeError:
            # 2.3 compat
            exc_info = self._exc_info_to_string(err)
        for cls, (storage, label, isfail) in self.errorClasses.items():
            if result.isclass(ec) and issubclass(ec, cls):
                if isfail:
                    test.passed = False
                storage.append((test, exc_info))
                # Might get patched into a streamless result
                if stream is not None:
                    if self.showAll:
                        message = [label]
                        detail = result._exception_detail(err[1])
                        if detail:
                            message.append(detail)
                        stream.writeln(": ".join(message))
                    elif self.dots:
                        stream.write(label[:1])
                return
        self.errors.append((test, exc_info))
        test.passed = False
        if stream is not None:
            self._writeResult(test, 'ERROR', 'red', 'E', False)

    @staticmethod
    def get_doc(cls_or_func):
        """Grabs the doc abbreviated doc string."""
        try:
            return cls_or_func.__doc__.split("\n")[0].strip()
        except (AttributeError, IndexError):
            return None

    def startTest(self, test):
        unittest.TestResult.startTest(self, test)
        self.start_time = time.time()
        test_name = None
        try:
            entry = test.test.__proboscis_case__.entry
            if entry.method:
                current_class = entry.method.im_class
                test_name = self.get_doc(entry.home) or entry.home.__name__
            else:
                current_class = entry.home
        except AttributeError:
            current_class = test.test.__class__

        if self.showAll:
            if current_class.__name__ != self._last_case:
                self.stream.writeln(current_class.__name__)
                self._last_case = current_class.__name__
                try:
                    doc = self.get_doc(current_class)
                except (AttributeError, IndexError):
                    doc = None
                if doc:
                    self.stream.writeln(' ' + doc)

            if not test_name:
                if hasattr(test.test, 'shortDescription'):
                    test_name = test.test.shortDescription()
                if not test_name:
                    test_name = test.test._testMethodName
            self.stream.write('\t%s' % str(test_name).ljust(60))
            self.stream.flush()


class NovaTestRunner(core.TextTestRunner):
    def __init__(self, *args, **kwargs):
        self.show_elapsed = kwargs.pop('show_elapsed')
        self.known_bugs = kwargs.pop('known_bugs', {})
        self.__result = None
        self.__finished = False
        self.__start_time = None
        super(NovaTestRunner, self).__init__(*args, **kwargs)

    def _makeResult(self):
        self.__result = NovaTestResult(
            self.stream,
            self.descriptions,
            self.verbosity,
            self.config,
            show_elapsed=self.show_elapsed,
            known_bugs=self.known_bugs)
        self.__start_time = time.time()
        return self.__result

    def _writeSlowTests(self, result_):
        # Pare out 'fast' tests
        slow_tests = [item for item in result_.slow_tests
                      if get_elapsed_time_color(item[0]) != 'green']
        if slow_tests:
            slow_total_time = sum(item[0] for item in slow_tests)
            self.stream.writeln("Slowest %i tests took %.2f secs:"
                                % (len(slow_tests), slow_total_time))
            for elapsed_time, test in sorted(slow_tests, reverse=True):
                time_str = "%.2f" % elapsed_time
                self.stream.writeln("    %s %s" % (time_str.ljust(10), test))

    def on_exit(self):
        if self.__result is None:
            print("Exiting before tests even started.")
        else:
            if not self.__finished:
                msg = "Tests aborted, trying to print available results..."
                print(msg)
                stop_time = time.time()
                self.__result.printErrors()
                self.__result.printSummary(self.__start_time, stop_time)
                self.config.plugins.finalize(self.__result)
                if self.show_elapsed:
                    self._writeSlowTests(self.__result)

    def run(self, test):
        result_ = super(NovaTestRunner, self).run(test)
        if self.show_elapsed:
            self._writeSlowTests(result_)
        self.__finished = True
        return result_


if __name__ == '__main__':
    logging.setup()
    # If any argument looks like a test name but doesn't have "nova.tests" in
    # front of it, automatically add that so we don't have to type as much
    show_elapsed = True
    argv = []
    test_fixture = os.getenv("UNITTEST_FIXTURE", "trove")
    for x in sys.argv:
        if x.startswith('test_'):
            argv.append('%s.tests.%s' % (test_fixture, x))
        elif x.startswith('--hide-elapsed'):
            show_elapsed = False
        else:
            argv.append(x)

    testdir = os.path.abspath(os.path.join(test_fixture, "tests"))
    c = config.Config(stream=sys.stdout,
                      env=os.environ,
                      verbosity=3,
                      workingDir=testdir,
                      plugins=core.DefaultPluginManager())

    runner = NovaTestRunner(stream=c.stream,
                            verbosity=c.verbosity,
                            config=c,
                            show_elapsed=show_elapsed)
    sys.exit(not core.run(config=c, testRunner=runner, argv=argv))



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import random
import redis

from trove.tests.scenario.helpers.test_helper import TestHelper
from trove.tests.scenario.runners.test_runners import TestRunner


class RedisHelper(TestHelper):

    def __init__(self, expected_override_name, report):
        super(RedisHelper, self).__init__(expected_override_name, report)

        self.key_patterns = ['user_a:%s', 'user_b:%s']
        self.value_pattern = 'id:%s'
        self.label_value = 'value_set'

        self._ds_client_cache = dict()

    def get_helper_credentials_root(self):
        return {'name': '-', 'password': 'rootpass'}

    def get_client(self, host, *args, **kwargs):
        # We need to cache the Redis client in order to prevent Error 99
        # (Cannot assign requested address) when working with large data sets.
        # A new client may be created frequently due to how the redirection
        # works (see '_execute_with_redirection').
        # The old (now closed) connections however have to wait for about 60s
        # (TIME_WAIT) before the port can be released.
        # This is a feature of the operating system that helps it dealing with
        # packets that arrive after the connection is closed.
        #
        # NOTE(zhaochao): when connecting to Redis server with a password,
        # current cached client may not updated to use the same password,
        # connection_kwargs of the ConnectPool object should be checked,
        # if the new password is different, A new client instance will be
        # created.
        recreate_client = True

        # NOTE(zhaochao): Another problem about caching clients is,  when
        # the 'requirepass' paramter of Redis server is changed, already
        # connected client can still issue commands. If we want to make sure
        # old passwords cannot be used to connect to the server, cached
        # clients shouldn't be used, a new one should be created instead.
        # We cannot easily tell whether the 'requirepass' paramter is changed.
        # So we have to always recreate a client when a password is explicitly
        # specified. The cached client is only used when no password
        # specified(i.e. we're going to use the default password) and the
        # cached password is same as the default one.
        if (host in self._ds_client_cache and 'password' not in kwargs):
            default_password = self.get_helper_credentials()['password']
            cached_password = (self._ds_client_cache[host]
                               .connection_pool
                               .connection_kwargs.get('password'))
            if cached_password == default_password:
                recreate_client = False

        if recreate_client:
            self._ds_client_cache[host] = (
                self.create_client(host, *args, **kwargs))

        return self._ds_client_cache[host]

    def create_client(self, host, *args, **kwargs):
        user = self.get_helper_credentials()
        password = kwargs.get('password', user['password'])
        client = redis.StrictRedis(password=password, host=host)
        return client

    # Add data overrides
    # We use multiple keys to make the Redis backup take longer
    def add_actual_data(self, data_label, data_start, data_size, host,
                        *args, **kwargs):
        test_set = self._get_data_point(host, data_label, *args, **kwargs)
        if not test_set:
            for num in range(data_start, data_start + data_size):
                for key_pattern in self.key_patterns:
                    self._set_data_point(
                        host,
                        key_pattern % str(num), self.value_pattern % str(num),
                        *args, **kwargs)
            # now that the data is there, add the label
            self._set_data_point(
                host,
                data_label, self.label_value,
                *args, **kwargs)

    def _set_data_point(self, host, key, value, *args, **kwargs):
        def set_point(client, key, value):
            return client.set(key, value)

        self._execute_with_redirection(
            host, set_point, [key, value], *args, **kwargs)

    def _get_data_point(self, host, key, *args, **kwargs):
        def get_point(client, key):
            return client.get(key)

        return self._execute_with_redirection(
            host, get_point, [key], *args, **kwargs)

    def _execute_with_redirection(self, host, callback, callback_args,
                                  *args, **kwargs):
        """Redis clustering is a relatively new feature still not supported
        in a fully transparent way by all clients.
        The application itself is responsible for connecting to the right node
        when accessing a key in a Redis cluster instead.

        Clients may be redirected to other nodes by redirection errors:

            redis.exceptions.ResponseError: MOVED 10778 10.64.0.2:6379

        This method tries to execute a given callback on a given host.
        If it gets a redirection error it parses the new host from the response
        and issues the same callback on this new host.
        """
        client = self.get_client(host, *args, **kwargs)
        try:
            return callback(client, *callback_args)
        except redis.exceptions.ResponseError as ex:
            response = str(ex)
            if response:
                tokens = response.split()
                if tokens[0] == 'MOVED':
                    redirected_host = tokens[2].split(':')[0]
                    if redirected_host:
                        return self._execute_with_redirection(
                            redirected_host, callback, callback_args,
                            *args, **kwargs)
            raise ex

    # Remove data overrides
    # We use multiple keys to make the Redis backup take longer
    def remove_actual_data(self, data_label, data_start, data_size, host,
                           *args, **kwargs):
        test_set = self._get_data_point(host, data_label, *args, **kwargs)
        if test_set:
            for num in range(data_start, data_start + data_size):
                for key_pattern in self.key_patterns:
                    self._expire_data_point(host, key_pattern % str(num),
                                            *args, **kwargs)
            # now that the data is gone, remove the label
            self._expire_data_point(host, data_label, *args, **kwargs)

    def _expire_data_point(self, host, key, *args, **kwargs):
        def expire_point(client, key):
            return client.expire(key, 0)

        self._execute_with_redirection(
            host, expire_point, [key], *args, **kwargs)

    # Verify data overrides
    # We use multiple keys to make the Redis backup take longer
    def verify_actual_data(self, data_label, data_start, data_size, host,
                           *args, **kwargs):
        # make sure the data is there - tests edge cases and a random one
        self._verify_data_point(host, data_label, self.label_value,
                                *args, **kwargs)
        midway_num = data_start + int(data_size / 2)
        random_num = random.randint(data_start + 2,
                                    data_start + data_size - 3)
        for num in [data_start,
                    data_start + 1,
                    midway_num,
                    random_num,
                    data_start + data_size - 2,
                    data_start + data_size - 1]:
            for key_pattern in self.key_patterns:
                self._verify_data_point(host,
                                        key_pattern % num,
                                        self.value_pattern % num,
                                        *args, **kwargs)
        # negative tests
        for num in [data_start - 1,
                    data_start + data_size]:
            for key_pattern in self.key_patterns:
                self._verify_data_point(host, key_pattern % num, None,
                                        *args, **kwargs)

    def _verify_data_point(self, host, key, expected_value, *args, **kwargs):
        value = self._get_data_point(host, key, *args, **kwargs)
        TestRunner.assert_equal(expected_value, value,
                                "Unexpected value '%s' returned from Redis "
                                "key '%s'" % (value, key))

    def get_dynamic_group(self):
        return {'hz': 15}

    def get_non_dynamic_group(self):
        return {'databases': 24}

    def get_invalid_groups(self):
        return [{'hz': 600}, {'databases': -1}, {'databases': 'string_value'}]

    def ping(self, host, *args, **kwargs):
        try:
            client = self.get_client(host, *args, **kwargs)
            return client.ping()
        except Exception:
            return False



# Copyright 2014 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import abc
import uuid

from oslo_log import log as logging
from oslo_utils import netutils

from trove.common import cfg
from trove.common import exception
from trove.common import utils
from trove.common.db.mysql import models
from trove.guestagent.common import operating_system
from trove.guestagent.strategies.replication import base

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class MysqlReplicationBase(base.Replication):
    """Base class for MySql Replication strategies."""

    def get_master_ref(self, service, snapshot_info):
        master_ref = {
            'host': netutils.get_my_ipv4(),
            'port': service.get_port()
        }
        return master_ref

    def _create_replication_user(self, service, adm):
        replication_user = None
        replication_password = utils.generate_random_password(16)

        mysql_user = None  # cache the model as we just want name validation

        retry_count = 0

        while replication_user is None:
            try:
                name = 'slave_' + str(uuid.uuid4())[:8]
                if mysql_user:
                    mysql_user.name = name
                else:
                    mysql_user = models.MySQLUser(
                        name=name, password=replication_password
                    )
                    mysql_user.check_create()

                LOG.debug("Trying to create replication user %s",
                          mysql_user.name)
                adm.create_users([mysql_user.serialize()])

                replication_user = {
                    'name': mysql_user.name,
                    'password': replication_password
                }
            except Exception:
                retry_count += 1
                if retry_count > 5:
                    LOG.error("Replication user retry count exceeded")
                    raise

        return replication_user

    def snapshot_for_replication(self, context, service, adm, location,
                                 snapshot_info):
        LOG.info("Creating backup for replication")

        volumes_mapping = {
            '/var/lib/mysql': {'bind': '/var/lib/mysql', 'mode': 'rw'},
            "/var/run/mysqld": {"bind": "/var/run/mysqld", "mode": "ro"},
            '/tmp': {'bind': '/tmp', 'mode': 'rw'}
        }
        service.create_backup(context, snapshot_info,
                              volumes_mapping=volumes_mapping)

        LOG.info('Creating replication user')
        replication_user = self._create_replication_user(service, adm)
        service.grant_replication_privilege(replication_user)

        replica_conf = {
            'log_position': {},
            'replication_user': replication_user
        }
        return snapshot_info['id'], replica_conf

    def enable_as_master(self, service, master_config):
        if not service.exists_replication_source_overrides():
            service.write_replication_source_overrides(master_config)
            service.restart()

    def read_last_master_gtid(self, service):
        INFO_FILE = ('%s/xtrabackup_binlog_info' % service.get_data_dir())
        operating_system.chmod(INFO_FILE,
                               operating_system.FileMode.ADD_READ_ALL,
                               as_root=True)

        LOG.info("Reading last master GTID from %s", INFO_FILE)
        try:
            with open(INFO_FILE, 'r') as f:
                content = f.read()
                LOG.debug('Content in %s: "%s"', INFO_FILE, content)
                ret = content.strip().split('\t')
                return ret[2] if len(ret) == 3 else ''
        except Exception as ex:
            LOG.error('Failed to read last master GTID, error: %s', str(ex))
            raise exception.UnableToDetermineLastMasterGTID(
                {'binlog_file': INFO_FILE})

    @abc.abstractmethod
    def connect_to_master(self, service, master_info):
        """Connects a slave to a master"""

    def enable_as_slave(self, service, master_info, slave_config):
        try:
            service.write_replication_replica_overrides(slave_config)
            service.restart()
            self.connect_to_master(service, master_info)
        except Exception as err:
            LOG.error("Exception enabling guest as replica, error: %s",
                      str(err))
            raise

    def detach_slave(self, service, for_failover):
        replica_info = service.stop_slave(for_failover)
        service.remove_replication_replica_overrides()
        service.restart()
        return replica_info

    def get_replica_context(self, service, adm):
        """Get replication information as master."""
        replication_user = self._create_replication_user(service, adm)
        service.grant_replication_privilege(replication_user)
        return {
            'master': self.get_master_ref(service, None),
            'replica_conf': {
                'log_position': {},
                'replication_user': replication_user
            }
        }

    def cleanup_source_on_replica_detach(self, admin_service, replica_info):
        admin_service.delete_user_by_name(replica_info['replication_user'])

    def demote_master(self, service):
        service.remove_replication_source_overrides()
        service.restart()



# Copyright 2014 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.common.clients import guest_client


class UpgradeMessageSender(object):
    """
    This class handles the business logic for sending
    an rpc message to the guest
    """

    @staticmethod
    def create(context, instance_id, instance_version, location,
               metadata=None):

        instance_id = UpgradeMessageSender._validate(instance_id, 36)

        if instance_version:
            instance_version = UpgradeMessageSender._validate(
                instance_version, 255)

        if location:
            location = UpgradeMessageSender._validate(location, 255)

        def _create_resources():
            guest_client(context, instance_id).upgrade(
                instance_version, location, metadata)
        return _create_resources

    @staticmethod
    def _validate(s, max_length):
        if s is None:
            raise ValueError()
        s = s.strip()
        length = len(s)
        if length < 1 or length > max_length:
            raise ValueError()
        return s



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from proboscis import test

from trove.tests.scenario import groups
from trove.tests.scenario.groups.test_group import TestGroup
from trove.tests.scenario.runners import test_runners


GROUP = "scenario.instance_actions_group"


class InstanceActionsRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'instance_actions_runners'
    _runner_cls = 'InstanceActionsRunner'


@test(depends_on_groups=[groups.INST_LOG],
      groups=[GROUP, groups.INST_ACTIONS])
class InstanceActionsGroup(TestGroup):
    """Test Instance Actions functionality."""

    def __init__(self):
        super(InstanceActionsGroup, self).__init__(
            InstanceActionsRunnerFactory.instance())

    @test
    def add_test_data(self):
        """Add test data."""
        self.test_runner.run_add_test_data()

    @test(depends_on=[add_test_data])
    def verify_test_data(self):
        """Verify test data."""
        self.test_runner.run_verify_test_data()

    @test(runs_after=[verify_test_data])
    def instance_restart(self):
        """Restart an existing instance."""
        self.test_runner.run_instance_restart()

    @test(depends_on=[verify_test_data, instance_restart])
    def verify_test_data_after_restart(self):
        """Verify test data after restart."""
        self.test_runner.run_verify_test_data()

    @test(depends_on=[instance_restart],
          runs_after=[verify_test_data_after_restart])
    def instance_resize_volume(self):
        """Resize attached volume."""
        self.test_runner.run_instance_resize_volume()

    @test(depends_on=[verify_test_data, instance_resize_volume])
    def verify_test_data_after_volume_resize(self):
        """Verify test data after volume resize."""
        self.test_runner.run_verify_test_data()

    @test(depends_on=[add_test_data],
          runs_after=[verify_test_data_after_volume_resize])
    def remove_test_data(self):
        """Remove test data."""
        self.test_runner.run_remove_test_data()


@test(depends_on_classes=[InstanceActionsGroup],
      groups=[GROUP, groups.INST_ACTIONS_RESIZE])
class InstanceActionsResizeGroup(TestGroup):
    """Test Instance Actions Resize functionality."""

    def __init__(self):
        super(InstanceActionsResizeGroup, self).__init__(
            InstanceActionsRunnerFactory.instance())

    @test
    def add_test_data(self):
        """Add test data."""
        self.test_runner.run_add_test_data()

    @test(depends_on=[add_test_data])
    def verify_test_data(self):
        """Verify test data."""
        self.test_runner.run_verify_test_data()

    @test(runs_after=[verify_test_data])
    def instance_resize_flavor(self):
        """Resize instance flavor."""
        self.test_runner.run_instance_resize_flavor()


@test(depends_on_classes=[InstanceActionsResizeGroup],
      groups=[GROUP, groups.INST_ACTIONS_RESIZE_WAIT])
class InstanceActionsResizeWaitGroup(TestGroup):
    """Test that Instance Actions Resize Completes."""

    def __init__(self):
        super(InstanceActionsResizeWaitGroup, self).__init__(
            InstanceActionsRunnerFactory.instance())

    @test
    def wait_for_instance_resize_flavor(self):
        """Wait for resize instance flavor to complete."""
        self.test_runner.run_wait_for_instance_resize_flavor()

    @test(depends_on=[wait_for_instance_resize_flavor])
    def verify_test_data_after_flavor_resize(self):
        """Verify test data after flavor resize."""
        self.test_runner.run_verify_test_data()

    @test(runs_after=[verify_test_data_after_flavor_resize])
    def remove_test_data(self):
        """Remove test data."""
        self.test_runner.run_remove_test_data()



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import traceback


class DefaultRootHandler(logging.StreamHandler):
    """A singleton StreamHandler"""
    __handler = logging.StreamHandler()
    __singleton = None
    __info = None
    __enable_backtrace = False

    @classmethod
    def activate(cls, enable_backtrace=False):
        # leverage the singleton __handler which has an
        # acquire() method to create a critical section.
        cls.__handler.acquire()
        if cls.__singleton is None:
            cls.__singleton = DefaultRootHandler()

        cls.__enable_backtrace = enable_backtrace
        cls.__handler.release()
        return cls.__singleton

    @classmethod
    def set_info(cls, info=None):
        cls.__info = info

    def __init__(self):
        if DefaultRootHandler.__singleton is not None:
            raise Exception(
                "Do not directly instantiate DefaultRootHandler(). "
                "Only use the activate() class method.")

        super(DefaultRootHandler, self).__init__()

    def emit(self, record):
        if DefaultRootHandler.__info:
            msg = ("*************************\n" +
                   "Unhandled message logged from " +
                   DefaultRootHandler.__info + ", " +
                   record.name + "\n")

            if DefaultRootHandler.__enable_backtrace:
                msg += ''.join(traceback.format_stack()) + "\n"

            msg += "*************************\n"
            self.stream.write(msg)
            self.flush()


class DefaultRootLogger(object):
    """A root logger that uses the singleton handler"""

    def __init__(self, enable_backtrace=False):
        super(DefaultRootLogger, self).__init__()
        handler = DefaultRootHandler.activate(
            enable_backtrace=enable_backtrace
        )

        handler.acquire()
        if handler not in logging.getLogger('').handlers:
            logging.getLogger('').addHandler(handler)

        handler.release()



# Copyright 2016 Tesora, Inc
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class VolumeTypeView(object):

    def __init__(self, volume_type, req=None):
        self.volume_type = volume_type
        self.req = req

    def data(self):
        volume_type = {
            'id': self.volume_type.id,
            'name': self.volume_type.name,
            'is_public': self.volume_type.is_public,
            'description': self.volume_type.description
        }
        return {"volume_type": volume_type}


class VolumeTypesView(object):

    def __init__(self, volume_types, req=None):
        self.volume_types = volume_types
        self.req = req

    def data(self):
        data = []
        for volume_type in self.volume_types:
            data.append(VolumeTypeView(volume_type,
                                       req=self.req).data()['volume_type'])

        return {"volume_types": data}



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from proboscis import SkipTest
from proboscis import test

from trove.tests.scenario import groups
from trove.tests.scenario.groups.test_group import TestGroup
from trove.tests.scenario.runners import test_runners


GROUP = "scenario.instance_upgrade_group"


class InstanceUpgradeRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'instance_upgrade_runners'
    _runner_cls = 'InstanceUpgradeRunner'


class UserActionsRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'user_actions_runners'
    _runner_cls = 'UserActionsRunner'


class DatabaseActionsRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'database_actions_runners'
    _runner_cls = 'DatabaseActionsRunner'


@test(depends_on_groups=[groups.INST_CREATE_WAIT],
      groups=[GROUP, groups.INST_UPGRADE],
      runs_after_groups=[groups.INST_ACTIONS])
class InstanceUpgradeGroup(TestGroup):

    def __init__(self):
        super(InstanceUpgradeGroup, self).__init__(
            InstanceUpgradeRunnerFactory.instance())
        self.database_actions_runner = DatabaseActionsRunnerFactory.instance()
        self.user_actions_runner = UserActionsRunnerFactory.instance()

    @test
    def create_user_databases(self):
        """Create user databases on an existing instance."""
        # These databases may be referenced by the users (below) so we need to
        # create them first.
        self.database_actions_runner.run_databases_create()

    @test(runs_after=[create_user_databases])
    def create_users(self):
        """Create users on an existing instance."""
        self.user_actions_runner.run_users_create()

    @test(runs_after=[create_users])
    def add_test_data(self):
        """Add test data."""
        self.test_runner.run_add_test_data()

    @test(depends_on=[add_test_data])
    def verify_test_data(self):
        """Verify test data."""
        self.test_runner.run_verify_test_data()

    @test(depends_on=[verify_test_data])
    def list_users_before_upgrade(self):
        """List the created users before upgrade."""
        self.user_actions_runner.run_users_list()

    @test(depends_on=[list_users_before_upgrade])
    def instance_upgrade(self):
        """Upgrade an existing instance."""
        raise SkipTest("Skip the instance upgrade integration test "
                       "temporarily because of not stable in CI")
        # self.test_runner.run_instance_upgrade()

    @test(depends_on=[list_users_before_upgrade])
    def show_user(self):
        """Show created users."""
        self.user_actions_runner.run_user_show()

    @test(depends_on=[create_users],
          runs_after=[show_user])
    def list_users(self):
        """List the created users."""
        self.user_actions_runner.run_users_list()

    @test(depends_on=[verify_test_data, instance_upgrade])
    def verify_test_data_after_upgrade(self):
        """Verify test data after upgrade."""
        self.test_runner.run_verify_test_data()

    @test(depends_on=[add_test_data],
          runs_after=[verify_test_data_after_upgrade])
    def remove_test_data(self):
        """Remove test data."""
        self.test_runner.run_remove_test_data()

    @test(depends_on=[create_users],
          runs_after=[list_users])
    def delete_user(self):
        """Delete the created users."""
        self.user_actions_runner.run_user_delete()

    @test(depends_on=[create_user_databases], runs_after=[delete_user])
    def delete_user_databases(self):
        """Delete the user databases."""
        self.database_actions_runner.run_database_delete()



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from sqlalchemy import ForeignKey
from sqlalchemy.schema import Column
from sqlalchemy.schema import MetaData

from trove.common import cfg
from trove.db.sqlalchemy.migrate_repo.schema import String
from trove.db.sqlalchemy.migrate_repo.schema import Table


CONF = cfg.CONF
logger = logging.getLogger('trove.db.sqlalchemy.migrate_repo.schema')

meta = MetaData()


def upgrade(migrate_engine):
    meta.bind = migrate_engine
    # Load 'configurations' table to MetaData.
    Table('configurations', meta, autoload=True, autoload_with=migrate_engine)
    instances = Table('clusters', meta, autoload=True)
    instances.create_column(Column('configuration_id', String(36),
                                   ForeignKey("configurations.id")))



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import abc
import functools
import re

from oslo_log import log as logging

from trove.common import exception


LOG = logging.getLogger(__name__)


class ModuleDriver(object, metaclass=abc.ABCMeta):
    """Base class that defines the contract for module drivers.

    Note that you don't have to derive from this class to have a valid
    driver; it is purely a convenience. Any class that adheres to the
    'interface' as dictated by this class' abstractmethod decorators
    (and other methods such as get_type, get_name and configure)
    will work.
    """
    def __init__(self):
        super(ModuleDriver, self).__init__()

        # This is used to store any message args to be substituted by
        # the output decorator when logging/returning messages.
        self._module_message_args = {}
        self._message_args = None
        self._generated_name = None

    @property
    def message_args(self):
        """Return a dict of message args that can be used to enhance
        the output decorator messages. This shouldn't be overridden; use
        self.message_args = <dict> instead to append values.
        """
        if not self._message_args:
            self._message_args = {
                'name': self.get_name(),
                'type': self.get_type()}
            self._message_args.update(self._module_message_args)
        return self._message_args

    @message_args.setter
    def message_args(self, values):
        """Set the message args that can be used to enhance
        the output decorator messages.
        """
        values = values or {}
        self._module_message_args = values
        self._message_args = None

    @property
    def generated_name(self):
        if not self._generated_name:
            # Turn class name into 'module type' format.
            # For example: DoCustomWorkDriver -> do_custom_work
            temp = re.sub('(.)[Dd]river$', r'\1', self.__class__.__name__)
            temp2 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', temp)
            temp3 = re.sub('([a-z0-9])([A-Z])', r'\1_\2', temp2)
            self._generated_name = temp3.lower()
        return self._generated_name

    def get_type(self):
        """This is used when setting up a module in Trove, and is here for
        code clarity.  It just returns the name of the driver by default.
        """
        return self.get_name()

    def get_name(self):
        """Use the generated name based on the class name. If
        overridden, must be in lower-case.
        """
        return self.generated_name

    @abc.abstractmethod
    def get_description(self):
        """Description for the driver."""
        pass

    @abc.abstractmethod
    def get_updated(self):
        """Date the driver was last updated."""
        pass

    @abc.abstractmethod
    def apply(self, name, datastore, ds_version, data_file, admin_module):
        """Apply the module to the guest instance. Return status and message
        as a tuple. Passes in whether the module was created with 'admin'
        privileges. This can be used as a form of access control by having
        the driver refuse to apply a module if it wasn't created with options
        that indicate that it was done by an 'admin' user.
        """
        return False, "Not a concrete driver"

    @abc.abstractmethod
    def remove(self, name, datastore, ds_version, data_file):
        """Remove the module from the guest instance.  Return
        status and message as a tuple.
        """
        return False, "Not a concrete driver"

    def configure(self, name, datastore, ds_version, data_file):
        """Configure the driver.  This is particularly useful for adding values
        to message_args, by having a line such as: self.message_args = <dict>.
        These values will be appended to the default ones defined
        in the message_args @property.
        """
        pass


def output(log_message=None, success_message=None,
           fail_message=None):
    """This is a decorator to trap the typical exceptions that occur
    when applying and removing modules. It returns the proper output
    corresponding to the error messages automatically. If the function
    returns output (success_flag, message) then those are returned,
    otherwise success is assumed and the success_message returned.
    Using this removes a lot of potential boiler-plate code, however
    it is not necessary.
    Keyword arguments can be used in the message string. Default
    values can be found in the message_args @property, however a
    driver can add whatever it see fit, by setting message_args
    to a dict in the configure call (see above). Thus if you set
    self.message_args = {'my_key': 'my_key_val'} then the message
    string could look like "My key is '$(my_key)s'".
    """
    success_message = success_message or "Success"
    fail_message = fail_message or "Fail"

    def output_decorator(func):
        """This is the actual decorator."""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            """Here's where we handle the error messages and return values
            from the actual function.
            """
            log_msg = log_message
            success_msg = success_message
            fail_msg = fail_message
            if isinstance(args[0], ModuleDriver):
                # Try and insert any message args if they exist in the driver
                message_args = args[0].message_args
                if message_args:
                    try:
                        log_msg = log_msg % message_args
                        success_msg = success_msg % message_args
                        fail_msg = fail_msg % message_args
                    except Exception:
                        # if there's a problem, just log it and drive on
                        LOG.warning("Could not apply message args: %s",
                                    message_args)
                        pass

            if log_msg:
                LOG.info(log_msg)
            success = False
            try:
                rv = func(*args, **kwargs)
                if rv:
                    # Use the actual values, if there are some
                    success, message = rv
                else:
                    success = True
                    message = success_msg
            except exception.ProcessExecutionError as ex:
                message = ("%(msg)s: %(out)s\n%(err)s" %
                           {'msg': fail_msg,
                            'out': ex.stdout,
                            'err': ex.stderr})
                message = message.replace(': \n', ': ')
                message = message.rstrip()
                LOG.exception(message)
            except exception.TroveError as ex:
                message = ("%(msg)s: %(err)s" %
                           {'msg': fail_msg, 'err': ex._error_string})
                LOG.exception(message)
            except Exception as ex:
                message = ("%(msg)s: %(err)s" %
                           {'msg': fail_msg, 'err': str(ex)})
                LOG.exception(message)
            return success, message

        return wrapper

    return output_decorator



# Copyright (c) 2013 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from collections import defaultdict

from oslo_log import log as logging
import proboscis.asserts as asserts
from proboscis.dependencies import SkipTest

from trove.common import utils
from trove.tests.config import CONFIG

LOG = logging.getLogger(__name__)
MESSAGE_QUEUE = defaultdict(list)


def create_usage_verifier():
    return utils.import_object(CONFIG.usage_endpoint)


class UsageVerifier(object):

    def clear_events(self):
        """Hook that is called to allow endpoints to clean up."""
        pass

    def check_message(self, resource_id, event_type, **attrs):
        messages = utils.poll_until(lambda: self.get_messages(resource_id),
                                    lambda x: len(x) > 0, time_out=30)
        found = None
        for message in messages:
            if message['event_type'] == event_type:
                found = message
        asserts.assert_is_not_none(found,
                                   "No message type %s for resource %s" %
                                   (event_type, resource_id))
        with asserts.Check() as check:
            for key, value in attrs.items():
                check.equal(found[key], value)

    def get_messages(self, resource_id, expected_messages=None):
        global MESSAGE_QUEUE
        msgs = MESSAGE_QUEUE.get(resource_id, [])
        if expected_messages is not None:
            asserts.assert_equal(len(msgs), expected_messages)
        return msgs


class FakeVerifier(object):
    """This is the default handler in fake mode, it is basically a no-op."""

    def clear_events(self):
        pass

    def check_message(self, *args, **kwargs):
        raise SkipTest("Notifications not available")

    def get_messages(self, *args, **kwargs):
        pass


def notify(event_type, payload):
    """Simple test notify function which saves the messages to global list."""
    payload['event_type'] = event_type
    if 'instance_id' in payload and 'server_type' not in payload:
        LOG.debug('Received Usage Notification: %s', event_type)
        resource_id = payload['instance_id']
        global MESSAGE_QUEUE
        MESSAGE_QUEUE[resource_id].append(payload)
        LOG.debug('Message Queue for %(id)s now has %(msg_count)d messages',
                  {'id': resource_id,
                   'msg_count': len(MESSAGE_QUEUE[resource_id])})



# Copyright 2015 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import abc
import operator

import docker
from oslo_config import cfg as oslo_cfg
from oslo_log import log as logging
from oslo_service import periodic_task

from trove.common import cfg
from trove.common import exception
from trove.common.i18n import _
from trove.common.notification import EndNotification
from trove.guestagent import dbaas
from trove.guestagent import guest_log
from trove.guestagent import volume
from trove.guestagent.common import guestagent_utils
from trove.guestagent.common import operating_system
from trove.guestagent.common.operating_system import FileMode
from trove.guestagent.module import driver_manager
from trove.guestagent.module import module_manager
from trove.guestagent.strategies import replication as repl_strategy
from trove.instance import service_status

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


class Manager(periodic_task.PeriodicTasks):
    """This is the base class for all datastore managers.  Over time, common
    functionality should be pulled back here from the existing managers.
    """

    GUEST_LOG_TYPE_LABEL = 'type'
    GUEST_LOG_USER_LABEL = 'user'
    GUEST_LOG_FILE_LABEL = 'file'
    GUEST_LOG_SECTION_LABEL = 'section'
    GUEST_LOG_ENABLE_LABEL = 'enable'
    GUEST_LOG_DISABLE_LABEL = 'disable'
    GUEST_LOG_RESTART_LABEL = 'restart'

    GUEST_LOG_BASE_DIR = '/var/log/trove'
    GUEST_LOG_DATASTORE_DIRNAME = 'datastore'

    GUEST_LOG_DEFS_GUEST_LABEL = 'guest'
    GUEST_LOG_DEFS_GENERAL_LABEL = 'general'
    GUEST_LOG_DEFS_ERROR_LABEL = 'error'
    GUEST_LOG_DEFS_SLOW_QUERY_LABEL = 'slow_query'

    MODULE_APPLY_TO_ALL = module_manager.ModuleManager.MODULE_APPLY_TO_ALL

    _docker_client = None

    @property
    def docker_client(self):
        if self._docker_client:
            return self._docker_client

        self._docker_client = docker.from_env()
        return self._docker_client

    def __init__(self, manager_name):
        super(Manager, self).__init__(CONF)

        # Manager properties
        self.__manager_name = manager_name
        self.__manager = None
        self.__prepare_error = False

        # Guest log
        self._guest_log_context = None
        self._guest_log_loaded_context = None
        self._guest_log_cache = None
        self._guest_log_defs = None

        # Module
        self.module_driver_manager = driver_manager.ModuleDriverManager()

        # Drivers should implement
        self.adm = None
        self.app = None
        self.status = None

        if CONF.guest_agent.container_registry:
            try:
                self.docker_client.login(
                    CONF.guest_agent.container_registry_username,
                    CONF.guest_agent.container_registry_password,
                    email="",
                    registry=CONF.guest_agent.container_registry)
            except Exception as exc:
                raise exception.GuestError(f"Failed to login the container "
                                           f"registry, error: {str(exc)}")

    @property
    def manager_name(self):
        """This returns the passed-in name of the manager."""
        return self.__manager_name

    @property
    def manager(self):
        """This returns the name of the manager."""
        if not self.__manager:
            self.__manager = CONF.datastore_manager or self.__manager_name
        return self.__manager

    @property
    def prepare_error(self):
        return self.__prepare_error

    @prepare_error.setter
    def prepare_error(self, prepare_error):
        self.__prepare_error = prepare_error

    @property
    def configuration_manager(self):
        """If the datastore supports the new-style configuration manager,
        it should override this to return it.
        """
        return None

    @property
    def replication(self):
        """If the datastore supports replication, return an instance of
        the strategy.
        """
        try:
            return repl_strategy.get_instance(self.manager)
        except Exception as ex:
            LOG.warning("Cannot get replication instance for '%(manager)s': "
                        "%(msg)s", {'manager': self.manager, 'msg': str(ex)})

        return None

    @property
    def replication_strategy(self):
        """If the datastore supports replication, return the strategy."""
        try:
            return repl_strategy.get_strategy(self.manager)
        except Exception as ex:
            LOG.debug("Cannot get replication strategy for '%(manager)s': "
                      "%(msg)s", {'manager': self.manager, 'msg': str(ex)})

        return None

    @property
    def guestagent_log_defs(self):
        """These are log files that should be available on every Trove
        instance.  By definition, these should be of type LogType.SYS
        """
        log_dir = CONF.log_dir or '/var/log/trove/'
        log_file = CONF.log_file or 'trove-guestagent.log'
        guestagent_log = guestagent_utils.build_file_path(log_dir, log_file)
        return {
            self.GUEST_LOG_DEFS_GUEST_LABEL: {
                self.GUEST_LOG_TYPE_LABEL: guest_log.LogType.SYS,
                self.GUEST_LOG_USER_LABEL: None,
                self.GUEST_LOG_FILE_LABEL: guestagent_log,
            },
        }

    @property
    def guest_log_context(self):
        return self._guest_log_context

    @guest_log_context.setter
    def guest_log_context(self, context):
        self._guest_log_context = context

    @periodic_task.periodic_task
    def update_status(self, context):
        """Update the status of the trove instance."""
        if not self.status.is_installed:
            LOG.info("Database service is not installed, skip status check")
            return

        LOG.debug("Starting to check database service status")

        status = self.get_service_status()
        self.status.set_status(status)

    def get_service_status(self):
        return self.status.get_actual_db_status()

    def rpc_ping(self, context):
        LOG.debug("Responding to RPC ping.")
        return True

    #################
    # Instance related
    #################
    def prepare(self, context, packages, databases, memory_mb, users,
                device_path=None, mount_point=None, backup_info=None,
                config_contents=None, root_password=None, overrides=None,
                cluster_config=None, snapshot=None, modules=None,
                ds_version=None):
        """Set up datastore on a Guest Instance."""
        with EndNotification(context, instance_id=CONF.guest_id):
            self._prepare(context, packages, databases, memory_mb, users,
                          device_path, mount_point, backup_info,
                          config_contents, root_password, overrides,
                          cluster_config, snapshot, modules,
                          ds_version=ds_version)

    def _prepare(self, context, packages, databases, memory_mb, users,
                 device_path, mount_point, backup_info,
                 config_contents, root_password, overrides,
                 cluster_config, snapshot, modules, ds_version=None):
        LOG.info("Starting datastore prepare for '%s:%s'.", self.manager,
                 ds_version)
        self.status.begin_install()
        post_processing = True if cluster_config else False
        try:
            # Since all module handling is common, don't pass it down to the
            # individual 'do_prepare' methods.
            self.do_prepare(context, packages, databases, memory_mb,
                            users, device_path, mount_point, backup_info,
                            config_contents, root_password, overrides,
                            cluster_config, snapshot, ds_version=ds_version)
        except Exception as ex:
            self.prepare_error = True
            LOG.exception("Failed to prepare datastore: %s", ex)
            raise
        finally:
            LOG.info("Ending datastore prepare for '%s'.", self.manager)
            self.status.end_install(error_occurred=self.prepare_error,
                                    post_processing=post_processing)
        # At this point critical 'prepare' work is done and the instance
        # is now in the correct 'ACTIVE' 'INSTANCE_READY' or 'ERROR' state.
        # Of cource if an error has occurred, none of the code that follows
        # will run.
        LOG.info("Completed setup of '%s' datastore successfully.",
                 self.manager)

        # The following block performs additional instance initialization.
        # Failures will be recorded, but won't stop the provisioning
        # or change the instance state.
        try:
            if modules:
                LOG.info("Applying modules (called from 'prepare').")
                self.module_apply(context, modules)
                LOG.info('Module apply completed.')
        except Exception as ex:
            LOG.exception("An error occurred applying modules: "
                          "%s", str(ex))

        # The following block performs single-instance initialization.
        # Failures will be recorded, but won't stop the provisioning
        # or change the instance state.
        if not cluster_config:
            try:
                if databases:
                    LOG.info("Creating databases (called from 'prepare').")
                    self.create_database(context, databases)
                    LOG.info('Databases created successfully.')
            except Exception as ex:
                LOG.warning("An error occurred creating databases: %s",
                            str(ex))
            try:
                if users:
                    LOG.info("Creating users (called from 'prepare')")
                    self.create_user(context, users)
                    LOG.info('Users created successfully.')
            except Exception as ex:
                LOG.warning("An error occurred creating users: "
                            "%s", str(ex))

            # We only enable-root automatically if not restoring a backup
            # that may already have root enabled in which case we keep it
            # unchanged.
            if root_password and not backup_info:
                try:
                    LOG.info("Enabling root user (with password).")
                    self.enable_root_on_prepare(context, root_password)
                    LOG.info('Root enabled successfully.')
                except Exception as ex:
                    LOG.exception("An error occurred enabling root user: "
                                  "%s", str(ex))

        try:
            LOG.info("Starting post prepare for '%s' datastore.", self.manager)
            self.post_prepare(context, packages, databases, memory_mb,
                              users, device_path, mount_point, backup_info,
                              config_contents, root_password, overrides,
                              cluster_config, snapshot)
            LOG.info("Post prepare for '%s' datastore completed.",
                     self.manager)
        except Exception as ex:
            LOG.exception("An error occurred in post prepare: %s",
                          str(ex))
            raise

    @abc.abstractmethod
    def do_prepare(self, context, packages, databases, memory_mb, users,
                   device_path, mount_point, backup_info, config_contents,
                   root_password, overrides, cluster_config, snapshot,
                   ds_version=None):
        """This is called from prepare when the Trove instance first comes
        online.  'Prepare' is the first rpc message passed from the
        task manager.  do_prepare handles all the base configuration of
        the instance and is where the actual work is done.  Once this method
        completes, the datastore is considered either 'ready' for use (or
        for final connections to other datastores) or in an 'error' state,
        and the status is changed accordingly.  Each datastore must
        implement this method.
        """
        pass

    def post_prepare(self, context, packages, databases, memory_mb, users,
                     device_path, mount_point, backup_info, config_contents,
                     root_password, overrides, cluster_config, snapshot):
        """This is called after prepare has completed successfully.
        Processing done here should be limited to things that will not
        affect the actual 'running' status of the datastore (for example,
        creating databases and users, although these are now handled
        automatically).  Any exceptions are caught, logged and rethrown,
        however no status changes are made and the end-user will not be
        informed of the error.
        """
        LOG.info('No post_prepare work has been defined.')
        pass

    def start_db_with_conf_changes(self, context, config_contents, ds_version):
        """Start the database with given configuration.

        This method is called after resize.
        """
        self.app.start_db_with_conf_changes(config_contents, ds_version)

    def stop_db(self, context):
        self.app.stop_db()

    def restart(self, context):
        self.app.restart()

    def rebuild(self, context, ds_version, config_contents=None,
                config_overrides=None):
        raise exception.DatastoreOperationNotSupported(
            operation='rebuild', datastore=self.manager)

    def pre_upgrade(self, context):
        """Prepares the guest for upgrade, returning a dict to be passed
        to post_upgrade
        """
        return {}

    def upgrade(self, context, upgrade_info):
        """Upgrade the database."""
        pass

    def post_upgrade(self, context, upgrade_info):
        """Recovers the guest after the image is upgraded using information
        from the pre_upgrade step
        """
        pass

    #####################
    # File System related
    #####################
    def get_filesystem_stats(self, context, fs_path):
        """Gets the filesystem stats for the path given."""
        # TODO(peterstac) - note that fs_path is not used in this method.
        mount_point = CONF.get(self.manager).mount_point
        LOG.debug("Getting file system stats for '%s'", mount_point)
        return dbaas.get_filesystem_volume_stats(mount_point)

    def mount_volume(self, context, device_path=None, mount_point=None,
                     write_to_fstab=False):
        LOG.debug("Mounting the device %(path)s at the mount point "
                  "%(mount_point)s.", {'path': device_path,
                                       'mount_point': mount_point})
        device = volume.VolumeDevice(device_path)
        device.mount(mount_point, write_to_fstab=write_to_fstab)

    def unmount_volume(self, context, device_path=None, mount_point=None):
        LOG.debug("Unmounting the device %(path)s from the mount point "
                  "%(mount_point)s.", {'path': device_path,
                                       'mount_point': mount_point})
        device = volume.VolumeDevice(device_path)
        device.unmount(mount_point)

    def resize_fs(self, context, device_path=None, mount_point=None,
                  online=False):
        LOG.info(f"Resizing the filesystem at {mount_point}, online: {online}")
        device = volume.VolumeDevice(device_path)
        device.resize_fs(mount_point, online=online)

    ###############
    # Configuration
    ###############
    def reset_configuration(self, context, configuration):
        """Reset database base configuration.

        The default implementation should be sufficient if a
        configuration_manager is provided. Even if one is not, this
        method needs to be implemented to allow the rollback of
        flavor-resize on the guestagent side.
        """
        if self.configuration_manager:
            LOG.info("Resetting configuration.")
            config_contents = configuration['config_contents']
            self.configuration_manager.reset_configuration(config_contents)

    def apply_overrides_on_prepare(self, context, overrides):
        self.update_overrides(context, overrides)
        self.restart(context)

    def update_overrides(self, context, overrides, remove=False):
        LOG.info(f"Updating config options: {overrides}, remove={remove}")
        if remove:
            self.app.remove_overrides()
        self.app.update_overrides(overrides)

    def apply_overrides(self, context, overrides):
        raise exception.DatastoreOperationNotSupported(
            operation='apply_overrides', datastore=self.manager)

    #################
    # Cluster related
    #################
    def cluster_complete(self, context):
        LOG.info("Cluster creation complete, starting status checks.")
        self.status.end_install()

    #############
    # Log related
    #############
    def get_datastore_log_defs(self):
        """Any datastore-specific log files should be overridden in this dict
        by the corresponding Manager class.

        Format of a dict entry:

        'name_of_log': {self.GUEST_LOG_TYPE_LABEL:
                            Specified by the Enum in guest_log.LogType,
                        self.GUEST_LOG_USER_LABEL:
                            User that owns the file,
                        self.GUEST_LOG_FILE_LABEL:
                            Path on filesystem where the log resides,
                        self.GUEST_LOG_SECTION_LABEL:
                            Section where to put config (if ini style)
                        self.GUEST_LOG_ENABLE_LABEL: {
                            Dict of config_group settings to enable log},
                        self.GUEST_LOG_DISABLE_LABEL: {
                            Dict of config_group settings to disable log},

        See guestagent_log_defs for an example.
        """
        return {}

    def is_log_enabled(self, logname):
        return False

    def get_guest_log_defs(self):
        """Return all the guest log defs."""
        if not self._guest_log_defs:
            self._guest_log_defs = dict(self.get_datastore_log_defs())
            self._guest_log_defs.update(self.guestagent_log_defs)
        return self._guest_log_defs

    def get_guest_log_cache(self):
        """Make sure the guest_log_cache is loaded and return it."""
        self._refresh_guest_log_cache()
        return self._guest_log_cache

    def _refresh_guest_log_cache(self):
        if self._guest_log_cache:
            # Replace the context if it's changed
            if self._guest_log_loaded_context != self.guest_log_context:
                for log_name in self._guest_log_cache.keys():
                    self._guest_log_cache[log_name].context = (
                        self.guest_log_context)
        else:
            # Load the initial cache
            self._guest_log_cache = {}
            if self.guest_log_context:
                gl_defs = self.get_guest_log_defs()
                try:
                    exposed_logs = CONF.get(self.manager).get(
                        'guest_log_exposed_logs')
                except oslo_cfg.NoSuchOptError:
                    exposed_logs = ''
                LOG.debug("Available log defs: %s", ",".join(gl_defs.keys()))
                exposed_logs = exposed_logs.lower().replace(',', ' ').split()
                LOG.debug("Exposing log defs: %s", ",".join(exposed_logs))
                expose_all = 'all' in exposed_logs

                for log_name in gl_defs.keys():
                    gl_def = gl_defs[log_name]
                    exposed = expose_all or log_name in exposed_logs
                    guestlog = guest_log.GuestLog(
                        self.guest_log_context, log_name,
                        gl_def[self.GUEST_LOG_TYPE_LABEL],
                        gl_def[self.GUEST_LOG_USER_LABEL],
                        gl_def[self.GUEST_LOG_FILE_LABEL],
                        exposed)

                    if (gl_def[self.GUEST_LOG_TYPE_LABEL] ==
                        guest_log.LogType.USER):
                        guestlog.enabled = self.is_log_enabled(log_name)
                        guestlog.status = (guest_log.LogStatus.Enabled
                                           if guestlog.enabled
                                           else guest_log.LogStatus.Disabled)

                    self._guest_log_cache[log_name] = guestlog

        self._guest_log_loaded_context = self.guest_log_context

    def guest_log_list(self, context):
        LOG.info("Getting list of guest logs.")
        self.guest_log_context = context
        gl_cache = self.get_guest_log_cache()
        result = filter(None, [gl_cache[log_name].show()
                               if gl_cache[log_name].exposed else None
                               for log_name in gl_cache.keys()])
        return result

    def guest_log_action(self, context, log_name, enable, disable,
                         publish, discard):
        if enable and disable:
            raise exception.BadRequest("Cannot enable and disable log '%s'." %
                                       log_name)
        # Enable if we are publishing, unless told to disable
        if publish and not disable:
            enable = True
        LOG.info("Processing guest log '%(log)s' "
                 "(enable=%(en)s, disable=%(dis)s, "
                 "publish=%(pub)s, discard=%(disc)s).",
                 {'log': log_name, 'en': enable, 'dis': disable,
                  'pub': publish, 'disc': discard})

        self.guest_log_context = context
        gl_cache = self.get_guest_log_cache()

        if log_name in gl_cache:
            LOG.debug(f"Found log {log_name}, type={gl_cache[log_name].type}, "
                      f"enable={gl_cache[log_name].enabled}")

            # system log can only be published
            if ((gl_cache[log_name].type == guest_log.LogType.SYS) and
                    not publish):
                if enable or disable:
                    if enable:
                        action_text = "enable"
                    else:
                        action_text = "disable"
                    raise exception.BadRequest("Cannot %s a SYSTEM log ('%s')."
                                               % (action_text, log_name))

            if gl_cache[log_name].type == guest_log.LogType.USER:
                requires_change = (
                    (gl_cache[log_name].enabled and disable) or
                    (not gl_cache[log_name].enabled and enable))
                if requires_change:
                    self.guest_log_enable(context, log_name, disable)
                    gl_cache[log_name].enabled = enable
                    gl_cache[log_name].status = (
                        guest_log.LogStatus.Enabled
                        if enable
                        else guest_log.LogStatus.Disabled
                    )

            log_details = gl_cache[log_name].show()
            if discard:
                log_details = gl_cache[log_name].discard_log()
            if publish:
                log_details = gl_cache[log_name].publish_log()

            LOG.info("Details for log '%(log)s': %(det)s",
                     {'log': log_name, 'det': log_details})
            return log_details

        raise exception.NotFound("Log '%s' is not defined." % log_name)

    def guest_log_enable(self, context, log_name, disable):
        """This method can be overridden by datastore implementations to
        facilitate enabling and disabling USER type logs.  If the logs
        can be enabled with simple configuration group changes, however,
        the code here will probably suffice.
        Must return whether the datastore needs to be restarted in order for
        the logging to begin.
        """
        restart_required = False
        verb = ("Disabling" if disable else "Enabling")
        if self.configuration_manager:
            LOG.debug("%(verb)s log '%(log)s'", {'verb': verb,
                                                 'log': log_name})
            gl_def = self.get_guest_log_defs()[log_name]
            enable_cfg_label = "%s_%s_log" % (self.GUEST_LOG_ENABLE_LABEL,
                                              log_name)
            disable_cfg_label = "%s_%s_log" % (self.GUEST_LOG_DISABLE_LABEL,
                                               log_name)
            restart_required = gl_def.get(self.GUEST_LOG_RESTART_LABEL,
                                          restart_required)
            if disable:
                self._apply_log_overrides(
                    context, enable_cfg_label, disable_cfg_label,
                    gl_def.get(self.GUEST_LOG_DISABLE_LABEL),
                    gl_def.get(self.GUEST_LOG_SECTION_LABEL),
                    restart_required)
            else:
                self._apply_log_overrides(
                    context, disable_cfg_label, enable_cfg_label,
                    gl_def.get(self.GUEST_LOG_ENABLE_LABEL),
                    gl_def.get(self.GUEST_LOG_SECTION_LABEL),
                    restart_required)
        else:
            log_fmt = ("%(verb)s log '%(log)s' not supported - "
                       "no configuration manager defined!")
            exc_fmt = _("%(verb)s log '%(log)s' not supported - "
                        "no configuration manager defined!")
            msg_content = {'verb': verb, 'log': log_name}
            LOG.error(log_fmt, msg_content)
            raise exception.GuestError(
                original_message=(exc_fmt % msg_content))

        return restart_required

    def _apply_log_overrides(self, context, remove_label,
                             apply_label, cfg_values, section_label,
                             restart_required):
        self.configuration_manager.remove_system_override(
            change_id=remove_label)
        if cfg_values:
            config_man_values = cfg_values
            if section_label:
                config_man_values = {section_label: cfg_values}
            self.configuration_manager.apply_system_override(
                config_man_values, change_id=apply_label, pre_user=True)
        if restart_required:
            self.status.set_status(
                service_status.ServiceStatuses.RESTART_REQUIRED)
        else:
            self.apply_overrides(context, cfg_values)

    def get_log_status(self, label):
        self.configuration_manager.get_value(label)

    def build_log_file_name(self, log_name, owner, datastore_dir=None):
        """Build a log file name based on the log_name and make sure the
        directories exist and are accessible by owner.
        """
        if datastore_dir is None:
            base_dir = self.GUEST_LOG_BASE_DIR
            if not operating_system.exists(base_dir, is_directory=True):
                operating_system.ensure_directory(
                    base_dir, user=owner, group=owner, force=True,
                    as_root=True)
            datastore_dir = guestagent_utils.build_file_path(
                base_dir, self.GUEST_LOG_DATASTORE_DIRNAME)

        if not operating_system.exists(datastore_dir, is_directory=True):
            operating_system.ensure_directory(
                datastore_dir, user=owner, group=owner, force=True,
                as_root=True)
        log_file_name = guestagent_utils.build_file_path(
            datastore_dir, '%s-%s.log' % (self.manager, log_name))

        return self.validate_log_file(log_file_name, owner)

    def validate_log_file(self, log_file, owner):
        """Make sure the log file exists and is accessible by owner.
        """
        if not operating_system.exists(log_file, as_root=True):
            operating_system.write_file(log_file, '', as_root=True)

        operating_system.chown(log_file, user=owner, group=owner,
                               as_root=True)
        operating_system.chmod(log_file, FileMode.ADD_USR_RW_GRP_RW_OTH_R,
                               as_root=True)

        return log_file

    ################
    # Module related
    ################
    def module_list(self, context, include_contents=False):
        LOG.info("Getting list of modules.")
        results = module_manager.ModuleManager.read_module_results(
            is_admin=context.is_admin, include_contents=include_contents)
        LOG.info("Returning list of modules: %s", results)
        return results

    def module_apply(self, context, modules=None):
        LOG.info("Applying modules.")
        results = []
        modules = [data['module'] for data in modules]
        try:
            # make sure the modules are applied in the correct order
            modules.sort(key=operator.itemgetter('apply_order'))
            modules.sort(key=operator.itemgetter('priority_apply'),
                         reverse=True)
        except KeyError:
            # If we don't have ordering info then maybe we're running
            # a version of the module feature before ordering was
            # introduced.  In that case, since we don't have any
            # way to order the modules we should just continue.
            pass
        for module in modules:
            id = module.get('id', None)
            module_type = module.get('type', None)
            name = module.get('name', None)
            tenant = module.get('tenant', self.MODULE_APPLY_TO_ALL)
            datastore = module.get('datastore', self.MODULE_APPLY_TO_ALL)
            ds_version = module.get('datastore_version',
                                    self.MODULE_APPLY_TO_ALL)
            contents = module.get('contents', None)
            md5 = module.get('md5', None)
            auto_apply = module.get('auto_apply', True)
            visible = module.get('visible', True)
            is_admin = module.get('is_admin', None)
            if is_admin is None:
                # fall back to the old method of checking for an admin option
                is_admin = (tenant == self.MODULE_APPLY_TO_ALL or
                            not visible or
                            auto_apply)
            if not name:
                raise AttributeError(_("Module name not specified"))
            if not contents:
                raise AttributeError(_("Module contents not specified"))
            driver = self.module_driver_manager.get_driver(module_type)
            if not driver:
                raise exception.ModuleTypeNotFound(
                    _("No driver implemented for module type '%s'") %
                    module_type)
            if (datastore and datastore != self.MODULE_APPLY_TO_ALL and
                    datastore != CONF.datastore_manager):
                reason = (_("Module not valid for datastore %s") %
                          CONF.datastore_manager)
                raise exception.ModuleInvalid(reason=reason)
            result = module_manager.ModuleManager.apply_module(
                driver, module_type, name, tenant, datastore, ds_version,
                contents, id, md5, auto_apply, visible, is_admin)
            results.append(result)
        LOG.info("Returning list of modules: %s", results)
        return results

    def module_remove(self, context, module=None):
        LOG.info("Removing module.")
        module = module['module']
        id = module.get('id', None)
        module_type = module.get('type', None)
        name = module.get('name', None)
        datastore = module.get('datastore', None)
        ds_version = module.get('datastore_version', None)
        if not name:
            raise AttributeError(_("Module name not specified"))
        driver = self.module_driver_manager.get_driver(module_type)
        if not driver:
            raise exception.ModuleTypeNotFound(
                _("No driver implemented for module type '%s'") %
                module_type)
        module_manager.ModuleManager.remove_module(
            driver, module_type, id, name, datastore, ds_version)
        LOG.info("Deleted module: %s", name)

    ################
    # Backup and restore
    ################
    def create_backup(self, context, backup_info):
        """Create backup for the database.

        :param context: User context object.
        :param backup_info: a dictionary containing the db instance id of the
                            backup task, location, type, and other data.
        """
        pass

    def perform_restore(self, context, restore_location, backup_info):
        LOG.info("Starting to restore database from backup %s, "
                 "backup_info: %s", backup_info['id'], backup_info)

        if (backup_info["location"].endswith('.enc') and
                not CONF.backup_aes_cbc_key):
            self.status.set_status(service_status.ServiceStatuses.FAILED)
            raise exception.TroveError('Decryption key not configured for '
                                       'encrypted backup.')

        try:
            self.app.restore_backup(context, backup_info, restore_location)
        except Exception:
            LOG.error("Failed to restore from backup %s.", backup_info['id'])
            self.status.set_status(service_status.ServiceStatuses.FAILED)
            raise

        LOG.info("Finished restore data from backup %s", backup_info['id'])

    ################
    # Database and user management
    ################
    def create_database(self, context, databases):
        with EndNotification(context):
            return self.adm.create_databases(databases)

    def list_databases(self, context, limit=None, marker=None,
                       include_marker=False):
        return self.adm.list_databases(limit, marker, include_marker)

    def delete_database(self, context, database):
        with EndNotification(context):
            return self.adm.delete_database(database)

    def change_passwords(self, context, users):
        with EndNotification(context):
            self.adm.change_passwords(users)

    def get_root_password(self, context):
        raise exception.DatastoreOperationNotSupported(
            operation='get_root_password', datastore=self.manager)

    def enable_root(self, context):
        LOG.info("Enabling root for the database.")
        return self.adm.enable_root()

    def enable_root_on_prepare(self, context, root_password):
        self.enable_root_with_password(context, root_password)

    def enable_root_with_password(self, context, root_password=None):
        return self.adm.enable_root(root_password)

    def disable_root(self, context):
        LOG.info("Disabling root for the database.")
        return self.adm.disable_root()

    def is_root_enabled(self, context):
        return self.adm.is_root_enabled()

    def create_user(self, context, users):
        with EndNotification(context):
            self.adm.create_users(users)

    def list_users(self, context, limit=None, marker=None,
                   include_marker=False):
        return self.adm.list_users(limit, marker, include_marker)

    def delete_user(self, context, user):
        with EndNotification(context):
            self.adm.delete_user(user)

    def get_user(self, context, username, hostname):
        return self.adm.get_user(username, hostname)

    def update_attributes(self, context, username, hostname, user_attrs):
        with EndNotification(context):
            self.adm.update_attributes(username, hostname, user_attrs)

    def grant_access(self, context, username, hostname, databases):
        return self.adm.grant_access(username, hostname, databases)

    def revoke_access(self, context, username, hostname, database):
        return self.adm.revoke_access(username, hostname, database)

    def list_access(self, context, username, hostname):
        return self.adm.list_access(username, hostname)

    ################
    # Replication related
    ################
    def backup_required_for_replication(self, context):
        return self.replication.backup_required_for_replication()

    def get_replication_snapshot(self, context, snapshot_info,
                                 replica_source_config=None):
        LOG.info("Getting replication snapshot, snapshot_info: %s",
                 snapshot_info)

        self.replication.enable_as_master(self.app, replica_source_config)
        LOG.info('Enabled as replication master')

        snapshot_id, replica_conf = self.replication.snapshot_for_replication(
            context, self.app, self.adm, None, snapshot_info)

        volume_stats = self.get_filesystem_stats(context, None)

        replication_snapshot = {
            'dataset': {
                'datastore_manager': self.manager,
                'dataset_size': volume_stats.get('used', 0.0),
                'volume_size': volume_stats.get('total', 0.0),
                'snapshot_id': snapshot_id
            },
            'replication_strategy': self.replication_strategy,
            'master': self.replication.get_master_ref(self.app, snapshot_info),
            'replica_conf': replica_conf
        }

        return replication_snapshot

    def attach_replica(self, context, snapshot, slave_config, restart=False):
        raise exception.DatastoreOperationNotSupported(
            operation='attach_replication_slave', datastore=self.manager)

    def detach_replica(self, context, for_failover=False):
        """Running on replica, detach from the primary."""
        LOG.info("Detaching replica.")
        replica_info = self.replication.detach_slave(self.app, for_failover)
        return replica_info

    def get_replica_context(self, context):
        """Running on primary."""
        LOG.info("Getting replica context.")
        replica_info = self.replication.get_replica_context(self.app, self.adm)
        return replica_info

    def make_read_only(self, context, read_only):
        raise exception.DatastoreOperationNotSupported(
            operation='make_read_only', datastore=self.manager)

    def enable_as_master(self, context, replica_source_config):
        LOG.info("Enable as master")
        self.replication.enable_as_master(self.app, replica_source_config)

    def demote_replication_master(self, context):
        LOG.info("Demoting replication master.")
        self.replication.demote_master(self.app)

    def get_txn_count(self, context):
        LOG.debug("Getting transaction count.")
        raise exception.DatastoreOperationNotSupported(
            operation='get_txn_count', datastore=self.manager)

    def get_latest_txn_id(self, context):
        raise exception.DatastoreOperationNotSupported(
            operation='get_latest_txn_id', datastore=self.manager)

    def wait_for_txn(self, context, txn):
        raise exception.DatastoreOperationNotSupported(
            operation='wait_for_txn', datastore=self.manager)



# Copyright (c) 2011 OpenStack, LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from proboscis import test
from proboscis.asserts import fail
from tests.util.services import Service
from trove.tests.config import CONFIG


def dbaas_url():
    return str(CONFIG.values.get("dbaas_url"))


def nova_url():
    return str(CONFIG.values.get("nova_client")['url'])


class Daemon(object):
    """Starts a daemon."""

    def __init__(self, alternate_path=None, conf_file_name=None,
                 extra_cmds=None, service_path_root=None, service_path=None):
        # The path to the daemon bin if the other one doesn't work.
        self.alternate_path = alternate_path
        self.extra_cmds = extra_cmds or []
        # The name of a test config value which points to a conf file.
        self.conf_file_name = conf_file_name
        # The name of a test config value, which is inserted into the service_path.
        self.service_path_root = service_path_root
        # The first path to the daemon bin we try.
        self.service_path = service_path or "%s"

    def run(self):
        # Print out everything to make it
        print("Looking for config value %s..." % self.service_path_root)
        print(CONFIG.values[self.service_path_root])
        path = self.service_path % CONFIG.values[self.service_path_root]
        print("Path = %s" % path)
        if not os.path.exists(path):
            path = self.alternate_path
        if path is None:
            fail("Could not find path to %s" % self.service_path_root)
        conf_path = str(CONFIG.values[self.conf_file_name])
        cmds = CONFIG.python_cmd_list() + [path] + self.extra_cmds + \
               [conf_path]
        print("Running cmds: %s" % cmds)
        self.service = Service(cmds)
        if not self.service.is_service_alive():
            self.service.start()



# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from unittest import mock
import uuid

from trove.datastore import models
from trove.instance.models import InstanceServiceStatus
from trove.instance.models import InstanceStatus
from trove.instance.models import SimpleInstance
from trove.instance.service_status import ServiceStatuses
from trove.instance.tasks import InstanceTasks
from trove.tests.unittests import trove_testtools
from trove.tests.unittests.util import util


class FakeInstanceTask(object):

    def __init__(self):
        self.is_error = False
        self.action = None


class FakeDBInstance(object):

    def __init__(self):
        self.id = str(uuid.uuid4())
        self.deleted = False
        self.datastore_version_id = str(uuid.uuid4())
        self.server_status = "ACTIVE"
        self.task_status = FakeInstanceTask()


class BaseInstanceStatusTestCase(trove_testtools.TestCase):
    @classmethod
    def setUpClass(cls):
        util.init_db()
        cls.db_info = FakeDBInstance()
        cls.datastore = models.DBDatastore.create(
            id=str(uuid.uuid4()),
            name='mysql' + str(uuid.uuid4()),
            default_version_id=cls.db_info.datastore_version_id
        )
        cls.version = models.DBDatastoreVersion.create(
            id=cls.db_info.datastore_version_id,
            datastore_id=cls.datastore.id,
            name='5.7' + str(uuid.uuid4()),
            manager='mysql',
            image_id=str(uuid.uuid4()),
            active=1,
            packages="mysql-server-5.7"
        )
        super(BaseInstanceStatusTestCase, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        util.cleanup_db()
        super(BaseInstanceStatusTestCase, cls).tearDownClass()


class InstanceStatusTest(BaseInstanceStatusTestCase):
    def setUp(self):
        self.db_info.task_status = FakeInstanceTask()
        self.db_info.server_status = "ACTIVE"
        self.ds_status = InstanceServiceStatus(ServiceStatuses.HEALTHY)
        super(InstanceStatusTest, self).setUp()

    def test_task_status_error_reports_error(self):
        self.db_info.task_status.is_error = True
        instance = SimpleInstance('dummy context', self.db_info,
                                  self.ds_status)
        self.assertEqual(InstanceStatus.ERROR, instance.status)

    def test_task_status_action_building_reports_build(self):
        self.db_info.task_status.action = "BUILDING"
        instance = SimpleInstance('dummy context', self.db_info,
                                  self.ds_status)
        self.assertEqual(InstanceStatus.BUILD, instance.status)

    def test_task_status_action_rebooting_reports_reboot(self):
        self.db_info.task_status.action = "REBOOTING"
        instance = SimpleInstance('dummy context', self.db_info,
                                  self.ds_status)
        self.assertEqual(InstanceStatus.REBOOT, instance.status)

    def test_task_status_action_resizing_reports_resize(self):
        self.db_info.task_status.action = "RESIZING"
        instance = SimpleInstance('dummy context', self.db_info,
                                  self.ds_status)
        self.assertEqual(InstanceStatus.RESIZE, instance.status)

    def test_task_deleting_server_active(self):
        self.db_info.task_status.action = "DELETING"
        instance = SimpleInstance('dummy context', self.db_info,
                                  self.ds_status)
        self.assertEqual(InstanceStatus.SHUTDOWN, instance.status)

    def test_nova_server_build_reports_build(self):
        self.db_info.server_status = "BUILD"
        instance = SimpleInstance('dummy context', self.db_info,
                                  self.ds_status)
        self.assertEqual(InstanceStatus.BUILD, instance.status)

    def test_nova_server_error_reports_error(self):
        self.db_info.server_status = "ERROR"
        instance = SimpleInstance('dummy context', self.db_info,
                                  self.ds_status)
        self.assertEqual(InstanceStatus.ERROR, instance.status)

    def test_nova_server_reboot_reports_reboot(self):
        self.db_info.server_status = "REBOOT"
        instance = SimpleInstance('dummy context', self.db_info,
                                  self.ds_status)
        self.assertEqual(InstanceStatus.REBOOT, instance.status)

    def test_nova_server_resize_reports_resize(self):
        self.db_info.server_status = "RESIZE"
        instance = SimpleInstance('dummy context', self.db_info,
                                  self.ds_status)
        self.assertEqual(InstanceStatus.RESIZE, instance.status)

    def test_nova_server_verify_resize_reports_resize(self):
        self.db_info.server_status = "VERIFY_RESIZE"
        instance = SimpleInstance('dummy context', self.db_info,
                                  self.ds_status)
        self.assertEqual(InstanceStatus.RESIZE, instance.status)

    def test_operating_status_healthy(self):
        self.db_info.task_status = InstanceTasks.NONE
        instance = SimpleInstance(mock.MagicMock(), self.db_info,
                                  self.ds_status)
        self.assertEqual(repr(ServiceStatuses.HEALTHY),
                         instance.operating_status)

    def test_operating_status_task_not_none(self):
        self.db_info.task_status = InstanceTasks.RESIZING
        instance = SimpleInstance(mock.MagicMock(), self.db_info,
                                  self.ds_status)
        self.assertEqual("",
                         instance.operating_status)



# Copyright 2014 eBay Software Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config.cfg import NoSuchOptError
from oslo_log import log as logging

from trove.cluster import models
from trove.cluster import views
from trove.common import apischema
from trove.common import cfg
from trove.common import clients
from trove.common import exception
from trove.common import glance as common_glance
from trove.common.i18n import _
from trove.common import notification
from trove.common.notification import StartNotification
from trove.common import pagination
from trove.common import policy
from trove.common import utils
from trove.common import wsgi
from trove.datastore import models as datastore_models


CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class ClusterController(wsgi.Controller):

    """Controller for cluster functionality."""
    schemas = apischema.cluster.copy()

    @classmethod
    def authorize_cluster_action(cls, context, cluster_rule_name, cluster):
        policy.authorize_on_target(context, 'cluster:%s' % cluster_rule_name,
                                   {'tenant': cluster.tenant_id})

    @classmethod
    def get_action_schema(cls, body, action_schema):
        action_type = list(body.keys())[0]
        return action_schema.get(action_type, {})

    @classmethod
    def get_schema(cls, action, body):
        action_schema = super(ClusterController, cls).get_schema(action, body)
        if action == 'action':
            action_schema = cls.get_action_schema(body, action_schema)
        return action_schema

    def action(self, req, body, tenant_id, id):
        LOG.debug(("Committing Action Against Cluster for "
                   "Tenant '%(tenant_id)s'\n"
                   "req : '%(req)s'\n\nid : '%(id)s'\n\n"),
                  {"req": req, "id": id, "tenant_id": tenant_id})
        if not body:
            raise exception.BadRequest(_("Invalid request body."))

        if len(body) != 1:
            raise exception.BadRequest(_("Action request should have exactly"
                                         " one action specified in body"))
        context = req.environ[wsgi.CONTEXT_KEY]
        cluster = models.Cluster.load(context, id)
        if ('reset-status' in body and
                'force_delete' not in body['reset-status']):
            self.authorize_cluster_action(context, 'reset-status', cluster)
        elif ('reset-status' in body and
                'force_delete' in body['reset-status']):
            self.authorize_cluster_action(context, 'force_delete', cluster)
        else:
            self.authorize_cluster_action(context, 'action', cluster)
        cluster.action(context, req, *next(iter(body.items())))

        view = views.load_view(cluster, req=req, load_servers=False)
        wsgi_result = wsgi.Result(view.data(), 202)

        return wsgi_result

    def show(self, req, tenant_id, id):
        """Return a single cluster."""
        LOG.debug(("Showing a Cluster for Tenant '%(tenant_id)s'\n"
                   "req : '%(req)s'\n\nid : '%(id)s'\n\n"),
                  {"req": req, "id": id, "tenant_id": tenant_id})

        context = req.environ[wsgi.CONTEXT_KEY]
        cluster = models.Cluster.load(context, id)
        self.authorize_cluster_action(context, 'show', cluster)
        return wsgi.Result(views.load_view(cluster, req=req).data(), 200)

    def show_instance(self, req, tenant_id, cluster_id, instance_id):
        """Return a single instance belonging to a cluster."""
        LOG.debug(("Showing an Instance in a Cluster for Tenant "
                   "'%(tenant_id)s'\n"
                   "req : '%(req)s'\n\n"
                   "cluster_id : '%(cluster_id)s'\n\n"
                   "instance_id : '%(instance_id)s;\n\n"),
                  {"req": req, "tenant_id": tenant_id,
                   "cluster_id": cluster_id,
                   "instance_id": instance_id})

        context = req.environ[wsgi.CONTEXT_KEY]
        cluster = models.Cluster.load(context, cluster_id)
        self.authorize_cluster_action(context, 'show_instance', cluster)
        instance = models.Cluster.load_instance(context, cluster.id,
                                                instance_id)
        return wsgi.Result(views.ClusterInstanceDetailView(
            instance, req=req).data(), 200)

    def delete(self, req, tenant_id, id):
        """Delete a cluster."""
        LOG.debug(("Deleting a Cluster for Tenant '%(tenant_id)s'\n"
                   "req : '%(req)s'\n\nid : '%(id)s'\n\n"),
                  {"req": req, "id": id, "tenant_id": tenant_id})

        context = req.environ[wsgi.CONTEXT_KEY]
        cluster = models.Cluster.load(context, id)
        self.authorize_cluster_action(context, 'delete', cluster)
        context.notification = notification.DBaaSClusterDelete(context,
                                                               request=req)
        with StartNotification(context, cluster_id=id):
            cluster.delete()
        return wsgi.Result(None, 202)

    def index(self, req, tenant_id):
        """Return a list of clusters."""
        LOG.debug(("Showing a list of clusters for Tenant '%(tenant_id)s'\n"
                   "req : '%(req)s'\n\n"), {"req": req,
                                            "tenant_id": tenant_id})

        context = req.environ[wsgi.CONTEXT_KEY]

        # This theoretically allows the Admin tenant list clusters for
        # only one particular tenant as opposed to listing all clusters for
        # for all tenants.
        # * As far as I can tell this is the only call which actually uses the
        #   passed-in 'tenant_id' for anything.
        if not context.is_admin and context.project_id != tenant_id:
            raise exception.TroveOperationAuthError(
                tenant_id=context.project_id
            )

        # The rule checks that the currently authenticated tenant can perform
        # the 'cluster-list' action.
        policy.authorize_on_tenant(context, 'cluster:index')

        # load all clusters and instances for the tenant
        clusters, marker = models.Cluster.load_all(context, tenant_id)
        view = views.ClustersView(clusters, req=req)
        paged = pagination.SimplePaginatedDataView(req.url, 'clusters', view,
                                                   marker)
        return wsgi.Result(paged.data(), 200)

    def create(self, req, body, tenant_id):
        LOG.debug(("Creating a Cluster for Tenant '%(tenant_id)s'\n"
                   "req : '%(req)s'\n\nbody : '%(body)s'\n\n"),
                  {"tenant_id": tenant_id, "req": req, "body": body})

        context = req.environ[wsgi.CONTEXT_KEY]
        policy.authorize_on_tenant(context, 'cluster:create')

        name = body['cluster']['name']
        datastore_args = body['cluster'].get('datastore', {})
        datastore, datastore_version = (
            datastore_models.get_datastore_version(**datastore_args))

        # Since Victoria, guest agent uses docker.
        # Get image_id from glance if image_id in datastore_versions table
        # is NULL.
        image_id = None
        if not datastore_version.image_id:
            glance_client = clients.create_glance_client(context)
            image_id = common_glance.get_image_id(
                glance_client, datastore_version.image_id,
                datastore_version.image_tags)

        extended_properties = body['cluster'].get('extended_properties', {})

        try:
            clusters_enabled = (CONF.get(datastore_version.manager)
                                .get('cluster_support'))
        except NoSuchOptError:
            clusters_enabled = False

        if not clusters_enabled:
            raise exception.ClusterDatastoreNotSupported(
                datastore=datastore.name,
                datastore_version=datastore_version.name)

        nodes = body['cluster']['instances']
        instances = []
        for node in nodes:
            flavor_id = utils.get_id_from_href(node['flavorRef'])
            volume_size = volume_type = nics = availability_zone = None
            modules = None
            if 'volume' in node:
                volume_size = int(node['volume']['size'])
                volume_type = node['volume'].get('type')
            if 'nics' in node:
                nics = node['nics']
            if 'availability_zone' in node:
                availability_zone = node['availability_zone']
            if 'modules' in node:
                modules = node['modules']

            instances.append({"flavor_id": flavor_id,
                              "volume_size": volume_size,
                              "volume_type": volume_type,
                              "nics": nics,
                              "availability_zone": availability_zone,
                              'region_name': node.get('region_name'),
                              "modules": modules})

        locality = body['cluster'].get('locality')
        if locality:
            locality_domain = ['affinity', 'anti-affinity']
            locality_domain_msg = ("Invalid locality '%s'. "
                                   "Must be one of ['%s']" %
                                   (locality,
                                    "', '".join(locality_domain)))
            if locality not in locality_domain:
                raise exception.BadRequest(message=locality_domain_msg)

        configuration = body['cluster'].get('configuration')

        context.notification = notification.DBaaSClusterCreate(context,
                                                               request=req)
        with StartNotification(context, name=name, datastore=datastore.name,
                               datastore_version=datastore_version.name):
            cluster = models.Cluster.create(context, name, datastore,
                                            datastore_version, instances,
                                            extended_properties,
                                            locality, configuration,
                                            image_id)
        cluster.locality = locality
        view = views.load_view(cluster, req=req, load_servers=False)
        return wsgi.Result(view.data(), 200)



# -*- coding: utf-8 -*-
#

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.todo',
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'openstackdocstheme',
    'stevedore.sphinxext',
    'oslo_config.sphinxconfiggen',
    'oslo_config.sphinxext',
    'oslo_policy.sphinxpolicygen',
    'oslo_policy.sphinxext',
]

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/trove'
openstackdocs_bug_project = 'trove'
openstack_auto_name = False
openstackdocs_bug_tag = ''
html_theme = 'openstackdocs'

# sphinxcontrib.apidoc options
config_generator_config_file = (
    '../../tools/trove-config-generator.conf')
sample_config_basename = '_static/trove'

policy_generator_config_file = (
    '../../tools/trove-policy-generator.conf')
sample_policy_basename = '_static/trove'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
#source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Trove'
copyright = '2013, OpenStack Foundation'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
#today = ''
# Else, today_fmt is used as the format for a strftime call.
#today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`) to use for all documents
#default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
#show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['trove.']

# If true, keep warnings as "system message" paragraphs in the built documents.
#keep_warnings = False

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# html_theme_path = ["."]
# html_theme = '_theme'
# html_static_path = ['_static']

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
#html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
#html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
#html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
#html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
#html_additional_pages = {}

# If false, no module index is generated.
#html_domain_indices = True

# If false, no index is generated.
#html_use_index = True

# If true, the index is split into individual pages for each letter.
#html_split_index = False

# If true, links to the reST sources are added to the pages.
#html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
#html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
#html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
#html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
#html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = '%sdoc' % project


# -- Options for LaTeX output -------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual])
latex_documents = [
    (
        'index',
        '%s.tex' % project,
        '%s Documentation' % project,
        'OpenStack Foundation',
        'manual'
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
#latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
#latex_use_parts = False

# If true, show page references after internal links.
#latex_show_pagerefs = False

# If true, show URL addresses after external links.
#latex_show_urls = False

# Documents to append as an appendix to all manuals.
#latex_appendices = []

# If false, no module index is generated.
#latex_domain_indices = True


# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
# man_pages = [
#     (
#         'index',
#         '%s' % project,
#         '%s Documentation' % project,
#         'OpenStack Foundation',
#         1
#     ),
# ]

# If true, show URL addresses after external links.
#man_show_urls = False


# -- Options for Texinfo output -----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        'index',
        '%s' % project,
        '%s Documentation' % project,
        'OpenStack Foundation',
        '%s' % project,
        'Database as a service.',
        'Miscellaneous'
        'manual'
    ),
]

# Documents to append as an appendix to all manuals.
#texinfo_appendices = []

# If false, no module index is generated.
#texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
#texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
#texinfo_no_detailmenu = False


# -- Options for Epub output --------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = '%s' % project
epub_author = 'OpenStack Foundation'
epub_publisher = 'OpenStack Foundation'
epub_copyright = '2013, OpenStack Foundation'

# The language of the text. It defaults to the language option
# or en if the language is not set.
#epub_language = ''

# The scheme of the identifier. Typical schemes are ISBN or URL.
#epub_scheme = ''

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#epub_identifier = ''

# A unique identification for the text.
#epub_uid = ''

# A tuple containing the cover image and cover page html template filenames.
#epub_cover = ()

# A sequence of (type, uri, title) tuples for the guide element of content.opf.
#epub_guide = ()

# HTML files that should be inserted before the pages created by sphinx.
# The format is a list of tuples containing the path and title.
#epub_pre_files = []

# HTML files shat should be inserted after the pages created by sphinx.
# The format is a list of tuples containing the path and title.
#epub_post_files = []

# A list of files that should not be packed into the epub file.
#epub_exclude_files = []

# The depth of the table of contents in toc.ncx.
#epub_tocdepth = 3

# Allow duplicate toc entries.
#epub_tocdup = True

# Fix unsupported image types using the PIL.
#epub_fix_images = False

# Scale large images.
#epub_max_image_width = 0

# If 'no', URL addresses will not be shown.
#epub_show_urls = 'inline'

# If false, no index is generated.
#epub_use_index = True



#    Copyright (c) 2014 Rackspace Hosting
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.datastore import models as datastore_models
from trove.datastore.models import Capability
from trove.datastore.models import Datastore
from trove.datastore.models import DatastoreVersion
from trove.datastore.models import DatastoreVersionMetadata
from trove.datastore.models import DBCapabilityOverrides
from trove.tests.unittests import trove_testtools
from trove.tests.unittests.util import util


class TestDatastoreBase(trove_testtools.TestCase):
    @classmethod
    def setUpClass(cls):
        util.init_db()

        cls.ds_name = cls.random_name(name='test-datastore')
        cls.ds_version_name = cls.random_name(name='test-version')
        cls.capability_name = cls.random_name(name='root_on_create',
                                              prefix='TestDatastoreBase')
        cls.capability_desc = "Enables root on create"
        cls.capability_enabled = True
        cls.flavor_id = 1
        cls.volume_type = 'some-valid-volume-type'

        datastore_models.update_datastore(cls.ds_name, False)
        cls.datastore = Datastore.load(cls.ds_name)

        datastore_models.update_datastore_version(
            cls.ds_name, cls.ds_version_name, "mysql", "", "", "", True)
        cls.datastore_version = DatastoreVersion.load(cls.datastore,
                                                      cls.ds_version_name)
        cls.test_id = cls.datastore_version.id

        DatastoreVersionMetadata.add_datastore_version_flavor_association(
            cls.datastore_version.id, [cls.flavor_id])
        DatastoreVersionMetadata.add_datastore_version_volume_type_association(
            cls.datastore_version.id, [cls.volume_type])

        cls.cap1 = Capability.create(cls.capability_name,
                                     cls.capability_desc, True)
        cls.cap2 = Capability.create(
            cls.random_name(name='require_volume', prefix='TestDatastoreBase'),
            "Require external volume", True)
        cls.cap3 = Capability.create(
            cls.random_name(name='test_capability',
                            prefix='TestDatastoreBase'),
            "Test capability", False)

        super(TestDatastoreBase, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        capabilities_overridden = DBCapabilityOverrides.find_all(
            datastore_version_id=cls.test_id).all()
        for ce in capabilities_overridden:
            ce.delete()

        cls.cap1.delete()
        cls.cap2.delete()
        cls.cap3.delete()

        datastore_models.DBDatastoreVersionMetadata.find_by(
            datastore_version_id=cls.test_id).delete()
        cls.datastore_version.delete()
        cls.datastore.delete()

        super(TestDatastoreBase, cls).tearDownClass()

    def capability_name_filter(self, capabilities):
        new_capabilities = []
        for capability in capabilities:
            if 'TestDatastoreBase' in capability.name:
                new_capabilities.append(capability)
        return new_capabilities



# Copyright 2015 Tesora Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from trove.tests.scenario.helpers.sql_helper import SqlHelper


class PostgresqlHelper(SqlHelper):

    def __init__(self, expected_override_name, report, port=5432):
        super(PostgresqlHelper, self).__init__(expected_override_name, report,
                                               'postgresql', port=port)

    @property
    def test_schema(self):
        return 'public'

    def get_helper_credentials(self):
        # There must be a database with the same name as the user in order
        # for the user to be able to login.
        return {'name': 'lite', 'password': 'litepass', 'database': 'lite'}

    def get_helper_credentials_root(self):
        return {'name': 'postgres', 'password': 'rootpass'}

    def get_valid_database_definitions(self):
        return [{'name': 'db1'}, {'name': 'db2'}, {'name': 'db3'}]

    def get_valid_user_definitions(self):
        return [{'name': 'user1', 'password': 'password1', 'databases': []},
                {'name': 'user2', 'password': 'password1',
                 'databases': [{'name': 'db1'}]},
                {'name': 'user3', 'password': 'password1',
                 'databases': [{'name': 'db1'}, {'name': 'db2'}]}]

    def get_dynamic_group(self):
        return {'effective_cache_size': '528MB'}

    def get_non_dynamic_group(self):
        return {'max_connections': 113,
                'log_min_duration_statement': '257ms'}

    def get_invalid_groups(self):
        return [{'timezone': 997},
                {"vacuum_cost_delay": 'string_value'},
                {"standard_conforming_strings": 'string_value'}]

    def get_configuration_value(self, property_name, host, *args, **kwargs):
        client = self.get_client(host, *args, **kwargs)
        cmd = "SHOW %s;" % property_name
        row = client.execute(cmd).fetchone()
        return row[0]

    def get_exposed_user_log_names(self):
        return ['general']

    def log_enable_requires_restart(self):
        return True



# Copyright 2016 Tesora, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from proboscis import test

from trove.tests.scenario import groups
from trove.tests.scenario.groups.test_group import TestGroup
from trove.tests.scenario.runners import test_runners


GROUP = "scenario.module_group"


class ModuleRunnerFactory(test_runners.RunnerFactory):

    _runner_ns = 'module_runners'
    _runner_cls = 'ModuleRunner'


@test(groups=[GROUP, groups.MODULE_CREATE])
class ModuleCreateGroup(TestGroup):
    """Test Module Create functionality."""

    def __init__(self):
        super(ModuleCreateGroup, self).__init__(
            ModuleRunnerFactory.instance())

    @test
    def module_delete_existing(self):
        """Delete all previous test modules."""
        self.test_runner.run_module_delete_existing()

    @test
    def module_create_bad_type(self):
        """Ensure create module with invalid type fails."""
        self.test_runner.run_module_create_bad_type()

    @test
    def module_create_non_admin_auto(self):
        """Ensure create auto_apply module for non-admin fails."""
        self.test_runner.run_module_create_non_admin_auto()

    @test
    def module_create_non_admin_all_tenant(self):
        """Ensure create all tenant module for non-admin fails."""
        self.test_runner.run_module_create_non_admin_all_tenant()

    @test
    def module_create_non_admin_hidden(self):
        """Ensure create hidden module for non-admin fails."""
        self.test_runner.run_module_create_non_admin_hidden()

    @test
    def module_create_non_admin_priority(self):
        """Ensure create priority module for non-admin fails."""
        self.test_runner.run_module_create_non_admin_priority()

    @test
    def module_create_non_admin_no_full_access(self):
        """Ensure create no full access module for non-admin fails."""
        self.test_runner.run_module_create_non_admin_no_full_access()

    @test
    def module_create_full_access_with_admin_opt(self):
        """Ensure create full access module with admin opts fails."""
        self.test_runner.run_module_create_full_access_with_admin_opt()

    @test
    def module_create_bad_datastore(self):
        """Ensure create module with invalid datastore fails."""
        self.test_runner.run_module_create_bad_datastore()

    @test
    def module_create_bad_datastore_version(self):
        """Ensure create module with invalid datastore_version fails."""
        self.test_runner.run_module_create_bad_datastore_version()

    @test
    def module_create_missing_datastore(self):
        """Ensure create module with missing datastore fails."""
        self.test_runner.run_module_create_missing_datastore()

    @test(runs_after=[module_delete_existing])
    def module_create(self):
        """Check that create module works."""
        self.test_runner.run_module_create()

    @test(runs_after=[module_create])
    def module_create_for_update(self):
        """Check that create module for update works."""
        self.test_runner.run_module_create_for_update()

    @test(depends_on=[module_create])
    def module_create_dupe(self):
        """Ensure create with duplicate info fails."""
        self.test_runner.run_module_create_dupe()

    @test(depends_on=[module_create_for_update])
    def module_update_missing_datastore(self):
        """Ensure update module with missing datastore fails."""
        self.test_runner.run_module_update_missing_datastore()

    @test(runs_after=[module_create_for_update])
    def module_create_bin(self):
        """Check that create module with binary contents works."""
        self.test_runner.run_module_create_bin()

    @test(runs_after=[module_create_bin])
    def module_create_bin2(self):
        """Check that create module with other binary contents works."""
        self.test_runner.run_module_create_bin2()

    @test(depends_on=[module_create])
    def module_show(self):
        """Check that show module works."""
        self.test_runner.run_module_show()

    @test(depends_on=[module_create])
    def module_show_unauth_user(self):
        """Ensure that show module for unauth user fails."""
        self.test_runner.run_module_show_unauth_user()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2])
    def module_list(self):
        """Check that list modules works."""
        self.test_runner.run_module_list()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2])
    def module_list_unauth_user(self):
        """Ensure that list module for unauth user fails."""
        self.test_runner.run_module_list_unauth_user()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_list])
    def module_create_admin_all(self):
        """Check that create module works with all admin options."""
        self.test_runner.run_module_create_admin_all()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_create_admin_all])
    def module_create_admin_hidden(self):
        """Check that create module works with hidden option."""
        self.test_runner.run_module_create_admin_hidden()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_create_admin_hidden])
    def module_create_admin_auto(self):
        """Check that create module works with auto option."""
        self.test_runner.run_module_create_admin_auto()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_create_admin_auto])
    def module_create_admin_live_update(self):
        """Check that create module works with live-update option."""
        self.test_runner.run_module_create_admin_live_update()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_create_admin_live_update])
    def module_create_admin_priority_apply(self):
        """Check that create module works with priority-apply option."""
        self.test_runner.run_module_create_admin_priority_apply()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_create_admin_priority_apply])
    def module_create_datastore(self):
        """Check that create module with datastore works."""
        self.test_runner.run_module_create_datastore()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_create_datastore])
    def module_create_different_datastore(self):
        """Check that create module with different datastore works."""
        self.test_runner.run_module_create_different_datastore()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_create_different_datastore])
    def module_create_ds_version(self):
        """Check that create module with ds version works."""
        self.test_runner.run_module_create_ds_version()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_create_ds_version])
    def module_create_all_tenant(self):
        """Check that create 'all' tenants with datastore module works."""
        self.test_runner.run_module_create_all_tenant()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_create_all_tenant, module_list_unauth_user])
    def module_create_different_tenant(self):
        """Check that create with same name on different tenant works."""
        self.test_runner.run_module_create_different_tenant()

    @test(depends_on=[module_create, module_create_bin, module_create_bin2],
          runs_after=[module_create_different_tenant])
    def module_create_full_access(self):
        """Check that create by admin with full access works."""
        self.test_runner.run_module_create_full_access()

    @test(depends_on=[module_create_all_tenant],
          runs_after=[module_create_full_access])
    def module_full_access_toggle(self):
        """Check that toggling full access works."""
        self.test_runner.run_module_full_access_toggle()

    @test(depends_on=[module_create_all_tenant],
          runs_after=[module_full_access_toggle])
    def module_list_again(self):
        """Check that list modules skips invisible modules."""
        self.test_runner.run_module_list_again()

    @test(depends_on=[module_create_ds_version],
          runs_after=[module_list_again])
    def module_list_ds(self):
        """Check that list modules by datastore works."""
        self.test_runner.run_module_list_ds()

    @test(depends_on=[module_create_ds_version],
          runs_after=[module_list_ds])
    def module_list_ds_all(self):
        """Check that list modules by all datastores works."""
        self.test_runner.run_module_list_ds_all()

    @test(depends_on=[module_create_admin_hidden])
    def module_show_invisible(self):
        """Ensure that show invisible module for non-admin fails."""
        self.test_runner.run_module_show_invisible()

    @test(depends_on=[module_create_all_tenant],
          runs_after=[module_create_different_tenant])
    def module_list_admin(self):
        """Check that list modules for admin works."""
        self.test_runner.run_module_list_admin()

    @test(depends_on=[module_create],
          runs_after=[module_show])
    def module_update(self):
        """Check that update module works."""
        self.test_runner.run_module_update()

    @test(depends_on=[module_update])
    def module_update_same_contents(self):
        """Check that update module with same contents works."""
        self.test_runner.run_module_update_same_contents()

    @test(depends_on=[module_update],
          runs_after=[module_update_same_contents])
    def module_update_auto_toggle(self):
        """Check that update module works for auto apply toggle."""
        self.test_runner.run_module_update_auto_toggle()

    @test(depends_on=[module_update],
          runs_after=[module_update_auto_toggle])
    def module_update_all_tenant_toggle(self):
        """Check that update module works for all tenant toggle."""
        self.test_runner.run_module_update_all_tenant_toggle()

    @test(depends_on=[module_update],
          runs_after=[module_update_all_tenant_toggle])
    def module_update_invisible_toggle(self):
        """Check that update module works for invisible toggle."""
        self.test_runner.run_module_update_invisible_toggle()

    @test(depends_on=[module_update],
          runs_after=[module_update_invisible_toggle])
    def module_update_priority_toggle(self):
        """Check that update module works for priority toggle."""
        self.test_runner.run_module_update_priority_toggle()

    @test(depends_on=[module_update],
          runs_after=[module_update_priority_toggle])
    def module_update_unauth(self):
        """Ensure update module for unauth user fails."""
        self.test_runner.run_module_update_unauth()

    @test(depends_on=[module_update],
          runs_after=[module_update_priority_toggle])
    def module_update_non_admin_auto(self):
        """Ensure update module to auto_apply for non-admin fails."""
        self.test_runner.run_module_update_non_admin_auto()

    @test(depends_on=[module_update],
          runs_after=[module_update_priority_toggle])
    def module_update_non_admin_auto_off(self):
        """Ensure update module to auto_apply off for non-admin fails."""
        self.test_runner.run_module_update_non_admin_auto_off()

    @test(depends_on=[module_update],
          runs_after=[module_update_priority_toggle])
    def module_update_non_admin_auto_any(self):
        """Ensure any update module to auto_apply for non-admin fails."""
        self.test_runner.run_module_update_non_admin_auto_any()

    @test(depends_on=[module_update],
          runs_after=[module_update_priority_toggle])
    def module_update_non_admin_all_tenant(self):
        """Ensure update module to all tenant for non-admin fails."""
        self.test_runner.run_module_update_non_admin_all_tenant()

    @test(depends_on=[module_update],
          runs_after=[module_update_priority_toggle])
    def module_update_non_admin_all_tenant_off(self):
        """Ensure update module to all tenant off for non-admin fails."""
        self.test_runner.run_module_update_non_admin_all_tenant_off()

    @test(depends_on=[module_update],
          runs_after=[module_update_priority_toggle])
    def module_update_non_admin_all_tenant_any(self):
        """Ensure any update module to all tenant for non-admin fails."""
        self.test_runner.run_module_update_non_admin_all_tenant_any()

    @test(depends_on=[module_update],
          runs_after=[module_update_priority_toggle])
    def module_update_non_admin_invisible(self):
        """Ensure update module to invisible for non-admin fails."""
        self.test_runner.run_module_update_non_admin_invisible()

    @test(depends_on=[module_update],
          runs_after=[module_update_priority_toggle])
    def module_update_non_admin_invisible_off(self):
        """Ensure update module to invisible off for non-admin fails."""
        self.test_runner.run_module_update_non_admin_invisible_off()

    @test(depends_on=[module_update],
          runs_after=[module_update_priority_toggle])
    def module_update_non_admin_invisible_any(self):
        """Ensure any update module to invisible for non-admin fails."""
        self.test_runner.run_module_update_non_admin_invisible_any()


@test(depends_on_groups=[groups.INST_CREATE_WAIT, groups.MODULE_CREATE],
      runs_after_groups=[groups.INST_ERROR_DELETE, groups.INST_FORCE_DELETE],
      groups=[GROUP, groups.MODULE_INST, groups.MODULE_INST_CREATE])
class ModuleInstCreateGroup(TestGroup):
    """Test Module Instance Create functionality."""

    def __init__(self):
        super(ModuleInstCreateGroup, self).__init__(
            ModuleRunnerFactory.instance())

    @test
    def module_list_instance_empty(self):
        """Check that the instance has no modules associated."""
        self.test_runner.run_module_list_instance_empty()

    @test(runs_after=[module_list_instance_empty])
    def module_instances_empty(self):
        """Check that the module hasn't been applied to any instances."""
        self.test_runner.run_module_instances_empty()

    @test(runs_after=[module_instances_empty])
    def module_instance_count_empty(self):
        """Check that no instance count exists."""
        self.test_runner.run_module_instance_count_empty()

    @test(runs_after=[module_instance_count_empty])
    def module_query_empty(self):
        """Check that the instance has no modules applied."""
        self.test_runner.run_module_query_empty()

    @test(runs_after=[module_query_empty])
    def module_apply(self):
        """Check that module-apply works."""
        self.test_runner.run_module_apply()

    @test(runs_after=[module_apply])
    def module_apply_wrong_module(self):
        """Ensure that module-apply for wrong module fails."""
        self.test_runner.run_module_apply_wrong_module()

    @test(depends_on=[module_apply_wrong_module])
    def module_update_not_live(self):
        """Ensure updating a non live_update module fails."""
        self.test_runner.run_module_update_not_live()

    @test(depends_on=[module_apply],
          runs_after=[module_update_not_live])
    def module_list_instance_after_apply(self):
        """Check that the instance has the modules associated."""
        self.test_runner.run_module_list_instance_after_apply()

    @test(runs_after=[module_list_instance_after_apply])
    def module_apply_live_update(self):
        """Check that module-apply works for live_update."""
        self.test_runner.run_module_apply_live_update()

    @test(depends_on=[module_apply_live_update])
    def module_list_instance_after_apply_live(self):
        """Check that the instance has the right modules."""
        self.test_runner.run_module_list_instance_after_apply_live()

    @test(runs_after=[module_list_instance_after_apply_live])
    def module_instances_after_apply(self):
        """Check that the instance shows up in the list."""
        self.test_runner.run_module_instances_after_apply()

    @test(runs_after=[module_instances_after_apply])
    def module_instance_count_after_apply(self):
        """Check that the instance count is right after apply."""
        self.test_runner.run_module_instance_count_after_apply()

    @test(runs_after=[module_instance_count_after_apply])
    def module_query_after_apply(self):
        """Check that module-query works."""
        self.test_runner.run_module_query_after_apply()

    @test(runs_after=[module_query_after_apply])
    def module_update_live_update(self):
        """Check that update module works on 'live' applied module."""
        self.test_runner.run_module_update_live_update()

    @test(runs_after=[module_update_live_update])
    def module_apply_another(self):
        """Check that module-apply works for another module."""
        self.test_runner.run_module_apply_another()

    @test(depends_on=[module_apply_another])
    def module_list_instance_after_apply_another(self):
        """Check that the instance has the right modules again."""
        self.test_runner.run_module_list_instance_after_apply_another()

    @test(runs_after=[module_list_instance_after_apply_another])
    def module_instances_after_apply_another(self):
        """Check that the instance shows up in the list still."""
        self.test_runner.run_module_instances_after_apply_another()

    @test(runs_after=[module_instances_after_apply_another])
    def module_instance_count_after_apply_another(self):
        """Check that the instance count is right after another apply."""
        self.test_runner.run_module_instance_count_after_apply_another()

    @test(depends_on=[module_apply_another],
          runs_after=[module_instance_count_after_apply_another])
    def module_query_after_apply_another(self):
        """Check that module-query works after another apply."""
        self.test_runner.run_module_query_after_apply_another()

    @test(depends_on=[module_apply],
          runs_after=[module_query_after_apply_another])
    def create_inst_with_mods(self):
        """Check that creating an instance with modules works."""
        self.test_runner.run_create_inst_with_mods()

    @test(runs_after=[create_inst_with_mods])
    def create_inst_with_wrong_module(self):
        """Ensure that creating an inst with wrong ds mod fails."""
        self.test_runner.run_create_inst_with_wrong_module()

    @test(depends_on=[module_apply],
          runs_after=[create_inst_with_wrong_module])
    def module_delete_applied(self):
        """Ensure that deleting an applied module fails."""
        self.test_runner.run_module_delete_applied()

    @test(depends_on=[module_apply],
          runs_after=[module_delete_applied])
    def module_remove(self):
        """Check that module-remove works."""
        self.test_runner.run_module_remove()

    @test(depends_on=[module_remove])
    def module_query_after_remove(self):
        """Check that the instance has modules applied after remove."""
        self.test_runner.run_module_query_after_remove()

    @test(depends_on=[module_remove],
          runs_after=[module_query_after_remove])
    def module_update_after_remove(self):
        """Check that update module after remove works."""
        self.test_runner.run_module_update_after_remove()

    @test(depends_on=[module_remove],
          runs_after=[module_update_after_remove])
    def module_apply_another_again(self):
        """Check that module-apply another works a second time."""
        self.test_runner.run_module_apply_another()

    @test(depends_on=[module_apply],
          runs_after=[module_apply_another_again])
    def module_query_after_apply_another2(self):
        """Check that module-query works still."""
        self.test_runner.run_module_query_after_apply_another()

    @test(depends_on=[module_apply_another_again],
          runs_after=[module_query_after_apply_another2])
    def module_remove_again(self):
        """Check that module-remove works again."""
        self.test_runner.run_module_remove()

    @test(depends_on=[module_remove_again])
    def module_query_empty_after_again(self):
        """Check that the inst has right mod applied after 2nd remove."""
        self.test_runner.run_module_query_after_remove()

    @test(depends_on=[module_remove_again],
          runs_after=[module_query_empty_after_again])
    def module_update_after_remove_again(self):
        """Check that update module after remove again works."""
        self.test_runner.run_module_update_after_remove_again()


@test(depends_on_groups=[groups.MODULE_INST_CREATE],
      groups=[GROUP, groups.MODULE_INST, groups.MODULE_INST_CREATE_WAIT],
      runs_after_groups=[groups.INST_ACTIONS, groups.INST_UPGRADE])
class ModuleInstCreateWaitGroup(TestGroup):
    """Test that Module Instance Create Completes."""

    def __init__(self):
        super(ModuleInstCreateWaitGroup, self).__init__(
            ModuleRunnerFactory.instance())

    @test
    def wait_for_inst_with_mods(self):
        """Wait for create instance with modules to finish."""
        self.test_runner.run_wait_for_inst_with_mods()

    @test(depends_on=[wait_for_inst_with_mods])
    def module_query_after_inst_create(self):
        """Check that module-query works on new instance."""
        self.test_runner.run_module_query_after_inst_create()

    @test(depends_on=[wait_for_inst_with_mods],
          runs_after=[module_query_after_inst_create])
    def module_retrieve_after_inst_create(self):
        """Check that module-retrieve works on new instance."""
        self.test_runner.run_module_retrieve_after_inst_create()

    @test(depends_on=[wait_for_inst_with_mods],
          runs_after=[module_retrieve_after_inst_create])
    def module_query_after_inst_create_admin(self):
        """Check that module-query works for admin."""
        self.test_runner.run_module_query_after_inst_create_admin()

    @test(depends_on=[wait_for_inst_with_mods],
          runs_after=[module_query_after_inst_create_admin])
    def module_retrieve_after_inst_create_admin(self):
        """Check that module-retrieve works for admin."""
        self.test_runner.run_module_retrieve_after_inst_create_admin()

    @test(depends_on=[wait_for_inst_with_mods],
          runs_after=[module_retrieve_after_inst_create_admin])
    def module_delete_auto_applied(self):
        """Ensure that module-delete on auto-applied module fails."""
        self.test_runner.run_module_delete_auto_applied()

    @test(runs_after=[module_delete_auto_applied])
    def module_list_instance_after_mod_inst(self):
        """Check that the new instance has the right modules."""
        self.test_runner.run_module_list_instance_after_mod_inst()

    @test(runs_after=[module_list_instance_after_mod_inst])
    def module_instances_after_mod_inst(self):
        """Check that the new instance shows up in the list."""
        self.test_runner.run_module_instances_after_mod_inst()

    @test(runs_after=[module_instances_after_mod_inst])
    def module_instance_count_after_mod_inst(self):
        """Check that the new instance count is right."""
        self.test_runner.run_module_instance_count_after_mod_inst()

    @test(runs_after=[module_instance_count_after_mod_inst])
    def module_reapply_with_md5(self):
        """Check that module reapply with md5 works."""
        self.test_runner.run_module_reapply_with_md5()

    @test(runs_after=[module_reapply_with_md5])
    def module_reapply_with_md5_verify(self):
        """Verify the dates after md5 reapply (no-op)."""
        self.test_runner.run_module_reapply_with_md5_verify()

    @test(runs_after=[module_reapply_with_md5_verify])
    def module_list_instance_after_reapply_md5(self):
        """Check that the instance's modules haven't changed."""
        self.test_runner.run_module_list_instance_after_reapply_md5()

    @test(runs_after=[module_list_instance_after_reapply_md5])
    def module_instances_after_reapply_md5(self):
        """Check that the new instance still shows up in the list."""
        self.test_runner.run_module_instances_after_reapply_md5()

    @test(runs_after=[module_instances_after_reapply_md5])
    def module_instance_count_after_reapply_md5(self):
        """Check that the instance count hasn't changed."""
        self.test_runner.run_module_instance_count_after_reapply_md5()

    @test(runs_after=[module_instance_count_after_reapply_md5])
    def module_reapply_all(self):
        """Check that module reapply works."""
        self.test_runner.run_module_reapply_all()

    @test(runs_after=[module_reapply_all])
    def module_reapply_all_wait(self):
        """Wait for module reapply to complete."""
        self.test_runner.run_module_reapply_all_wait()

    @test(runs_after=[module_reapply_all_wait])
    def module_instance_count_after_reapply(self):
        """Check that the reapply instance count is right."""
        self.test_runner.run_module_instance_count_after_reapply()

    @test(runs_after=[module_instance_count_after_reapply])
    def module_reapply_with_force(self):
        """Check that module reapply with force works."""
        self.test_runner.run_module_reapply_with_force()

    @test(runs_after=[module_reapply_with_force])
    def module_reapply_with_force_wait(self):
        """Wait for module reapply with force to complete."""
        self.test_runner.run_module_reapply_with_force_wait()

    @test(runs_after=[module_reapply_with_force_wait])
    def module_list_instance_after_reapply_force(self):
        """Check that the new instance still has the right modules."""
        self.test_runner.run_module_list_instance_after_reapply()

    @test(runs_after=[module_list_instance_after_reapply_force])
    def module_instances_after_reapply_force(self):
        """Check that the new instance still shows up in the list."""
        self.test_runner.run_module_instances_after_reapply()

    @test(runs_after=[module_instances_after_reapply_force])
    def module_instance_count_after_reapply_force(self):
        """Check that the instance count is right after reapply force."""
        self.test_runner.run_module_instance_count_after_reapply()


@test(depends_on_groups=[groups.MODULE_INST_CREATE_WAIT],
      groups=[GROUP, groups.MODULE_INST, groups.MODULE_INST_DELETE])
class ModuleInstDeleteGroup(TestGroup):
    """Test Module Instance Delete functionality."""

    def __init__(self):
        super(ModuleInstDeleteGroup, self).__init__(
            ModuleRunnerFactory.instance())

    @test
    def delete_inst_with_mods(self):
        """Check that instance with module can be deleted."""
        self.test_runner.run_delete_inst_with_mods()

    @test(runs_after=[delete_inst_with_mods])
    def remove_mods_from_main_inst(self):
        """Check that modules can be removed from the main instance."""
        self.test_runner.run_remove_mods_from_main_inst()


@test(depends_on_groups=[groups.MODULE_INST_DELETE],
      groups=[GROUP, groups.MODULE_INST, groups.MODULE_INST_DELETE_WAIT],
      runs_after_groups=[groups.INST_DELETE])
class ModuleInstDeleteWaitGroup(TestGroup):
    """Test that Module Instance Delete Completes."""

    def __init__(self):
        super(ModuleInstDeleteWaitGroup, self).__init__(
            ModuleRunnerFactory.instance())

    @test
    def wait_for_delete_inst_with_mods(self):
        """Wait until the instance with module is gone."""
        self.test_runner.run_wait_for_delete_inst_with_mods()


@test(depends_on_groups=[groups.MODULE_CREATE],
      runs_after_groups=[groups.MODULE_INST_DELETE_WAIT],
      groups=[GROUP, groups.MODULE_DELETE])
class ModuleDeleteGroup(TestGroup):
    """Test Module Delete functionality."""

    def __init__(self):
        super(ModuleDeleteGroup, self).__init__(
            ModuleRunnerFactory.instance())

    def module_delete_non_existent(self):
        """Ensure delete non-existent module fails."""
        self.test_runner.run_module_delete_non_existent()

    def module_delete_unauth_user(self):
        """Ensure delete module by unauth user fails."""
        self.test_runner.run_module_delete_unauth_user()

    @test(runs_after=[module_delete_unauth_user,
                      module_delete_non_existent])
    def module_delete_hidden_by_non_admin(self):
        """Ensure delete hidden module by non-admin user fails."""
        self.test_runner.run_module_delete_hidden_by_non_admin()

    @test(runs_after=[module_delete_hidden_by_non_admin])
    def module_delete_all_tenant_by_non_admin(self):
        """Ensure delete all tenant module by non-admin user fails."""
        self.test_runner.run_module_delete_all_tenant_by_non_admin()

    @test(runs_after=[module_delete_all_tenant_by_non_admin])
    def module_delete_auto_by_non_admin(self):
        """Ensure delete auto-apply module by non-admin user fails."""
        self.test_runner.run_module_delete_auto_by_non_admin()

    @test(runs_after=[module_delete_auto_by_non_admin])
    def module_delete(self):
        """Check that delete module works."""
        self.test_runner.run_module_delete()

    @test(runs_after=[module_delete])
    def module_delete_admin(self):
        """Check that delete module works for admin."""
        self.test_runner.run_module_delete_admin()

    @test(runs_after=[module_delete_admin])
    def module_delete_remaining(self):
        """Delete all remaining test modules."""
        self.test_runner.run_module_delete_existing()
